import numbers
from math import pi
import numpy as np

import gpaw.cgpaw as cgpaw
import gpaw.fftw as fftw
from gpaw.utilities.blas import mmm, r2k, rk
from gpaw.gpu import cupy as cp


class PWDescriptor:
    ndim = 1  # all 3d G-vectors are stored in a 1d ndarray

    def __init__(self, ecut, gd, dtype=None, kd=None,
                 fftwflags=fftw.MEASURE, gammacentered=False):

        assert gd.pbc_c.all()

        self.gd = gd
        self.fftwflags = fftwflags

        N_c = gd.N_c
        self.comm = gd.comm

        ecut0 = 0.5 * pi**2 / (self.gd.h_cv**2).sum(1).max()
        if ecut is None:
            ecut = 0.9999 * ecut0
        else:
            assert ecut <= ecut0

        self.ecut = ecut

        if dtype is None:
            if kd is None or kd.gamma:
                dtype = float
            else:
                dtype = complex
        self.dtype = dtype
        self.gammacentered = gammacentered

        if dtype == float:
            Nr_c = N_c.copy()
            Nr_c[2] = N_c[2] // 2 + 1
            i_Qc = np.indices(Nr_c).transpose((1, 2, 3, 0))
            i_Qc[..., :2] += N_c[:2] // 2
            i_Qc[..., :2] %= N_c[:2]
            i_Qc[..., :2] -= N_c[:2] // 2
            self.tmp_Q = fftw.empty(Nr_c, complex)
            self.tmp_R = self.tmp_Q.view(float)[:, :, :N_c[2]]
        else:
            i_Qc = np.indices(N_c).transpose((1, 2, 3, 0))
            i_Qc += N_c // 2
            i_Qc %= N_c
            i_Qc -= N_c // 2
            self.tmp_Q = fftw.empty(N_c, complex)
            self.tmp_R = self.tmp_Q

        self.fftplan = fftw.create_plan(self.tmp_R, self.tmp_Q, -1, fftwflags)
        self.ifftplan = fftw.create_plan(self.tmp_Q, self.tmp_R, 1, fftwflags)

        # Calculate reciprocal lattice vectors:
        B_cv = 2.0 * pi * gd.icell_cv
        i_Qc.shape = (-1, 3)
        self.G_Qv = np.dot(i_Qc, B_cv)

        self.kd = kd
        if kd is None:
            self.K_qv = np.zeros((1, 3))
            self.only_one_k_point = True
        else:
            self.K_qv = np.dot(kd.ibzk_qc, B_cv)
            self.only_one_k_point = (kd.nbzkpts == 1)

        # Map from vectors inside sphere to fft grid:
        self.Q_qG = []
        G2_qG = []
        Q_Q = np.arange(len(i_Qc), dtype=np.int32)

        self.ng_q = []
        for q, K_v in enumerate(self.K_qv):
            G2_Q = ((self.G_Qv + K_v)**2).sum(axis=1)
            if gammacentered:
                mask_Q = ((self.G_Qv**2).sum(axis=1) <= 2 * ecut)
            else:
                mask_Q = (G2_Q <= 2 * ecut)

            if self.dtype == float:
                mask_Q &= ((i_Qc[:, 2] > 0) |
                           (i_Qc[:, 1] > 0) |
                           ((i_Qc[:, 0] >= 0) & (i_Qc[:, 1] == 0)))
            Q_G = Q_Q[mask_Q]
            self.Q_qG.append(Q_G)
            G2_qG.append(G2_Q[Q_G])
            ng = len(Q_G)
            self.ng_q.append(ng)

        self.ngmin = min(self.ng_q)
        self.ngmax = max(self.ng_q)

        if kd is not None:
            self.ngmin = kd.comm.min_scalar(self.ngmin)
            self.ngmax = kd.comm.max_scalar(self.ngmax)

        # Distribute things:
        S = gd.comm.size
        self.maxmyng = (self.ngmax + S - 1) // S
        ng1 = gd.comm.rank * self.maxmyng
        ng2 = ng1 + self.maxmyng

        self.G2_qG = []
        self.myQ_qG = []
        self.myng_q = []
        for q, G2_G in enumerate(G2_qG):
            G2_G = G2_G[ng1:ng2].copy()
            G2_G.flags.writeable = False
            self.G2_qG.append(G2_G)
            myQ_G = self.Q_qG[q][ng1:ng2]
            self.myQ_qG.append(myQ_G)
            self.myng_q.append(len(myQ_G))

        if S > 1:
            self.tmp_G = np.empty(self.maxmyng * S, complex)
        else:
            self.tmp_G = None

    def get_reciprocal_vectors(self, q=0, add_q=True):
        """Returns reciprocal lattice vectors plus q, G + q,
        in xyz coordinates."""

        if add_q:
            q_v = self.K_qv[q]
            return self.G_Qv[self.myQ_qG[q]] + q_v
        return self.G_Qv[self.myQ_qG[q]]

    def __getstate__(self):
        return (self.ecut, self.gd, self.dtype, self.kd, self.fftwflags)

    def __setstate__(self, state):
        self.__init__(*state)

    def estimate_memory(self, mem):
        nbytes = (self.tmp_R.nbytes +
                  self.G_Qv.nbytes +
                  len(self.K_qv) * (self.ngmax * 4 +
                                    self.maxmyng * (8 + 4)))
        mem.subnode('Arrays', nbytes)

    def bytecount(self, dtype=float):
        return self.ngmax * 16

    def zeros(self, x=(), dtype=None, q=None, global_array=False):
        """Return zeroed array.

        The shape of the array will be x + (ng,) where ng is the number
        of G-vectors for on this core.  Different k-points will have
        different values for ng.  Therefore, the q index must be given,
        unless we are describibg a real-valued function."""

        a_xG = self.empty(x, dtype, q, global_array)
        a_xG.fill(0.0)
        return a_xG

    def empty(self, x=(), dtype=None, q=None, global_array=False):
        """Return empty array."""
        if dtype is not None:
            assert dtype == self.dtype
        if isinstance(x, numbers.Integral):
            x = (x,)
        if q is None:
            assert self.only_one_k_point
            q = 0
        if global_array:
            shape = x + (self.ng_q[q],)
        else:
            shape = x + (self.myng_q[q],)
        return np.empty(shape, complex)

    def fft(self, f_R, q=None, Q_G=None, local=False):
        """Fast Fourier transform.

        Returns c(G) for G<Gc::

                   --      -iG.R
            c(G) = > f(R) e
                   --
                   R

        If local=True, all cores will do an FFT without any
        collect/scatter.
        """

        if local:
            self.tmp_R[:] = f_R
        else:
            self.gd.collect(f_R, self.tmp_R)

        if self.gd.comm.rank == 0 or local:
            self.fftplan.execute()
            if Q_G is None:
                q = q or 0
                Q_G = self.Q_qG[q]
            f_G = self.tmp_Q.ravel()[Q_G]
            if local:
                return f_G
        else:
            f_G = None

        return self.scatter(f_G, q)

    def ifft(self, c_G, q=None, local=False, safe=True, distribute=True):
        """Inverse fast Fourier transform.

        Returns::

                   1 --        iG.R
            f(R) = - > c(G) e
                   N --
                     G

        If local=True, all cores will do an iFFT without any
        gather/distribute.
        """
        assert q is not None or self.only_one_k_point
        q = q or 0
        if not local:
            c_G = self.gather(c_G, q)
        comm = self.gd.comm
        scale = 1.0 / self.tmp_R.size
        if comm.rank == 0 or local:
            # Same as:
            #
            #    self.tmp_Q[:] = 0.0
            #    self.tmp_Q.ravel()[self.Q_qG[q]] = scale * c_G
            #
            # but much faster:
            Q_G = self.Q_qG[q]
            assert len(c_G) == len(Q_G)
            cgpaw.pw_insert(c_G, Q_G, scale, self.tmp_Q)

            if self.dtype == float:
                t = self.tmp_Q[:, :, 0]
                n, m = self.gd.N_c[:2] // 2 - 1
                t[0, -m:] = t[0, m:0:-1].conj()
                t[n:0:-1, -m:] = t[-n:, m:0:-1].conj()
                t[-n:, -m:] = t[n:0:-1, m:0:-1].conj()
                t[-n:, 0] = t[n:0:-1, 0].conj()
            self.ifftplan.execute()
        if comm.size == 1 or local or not distribute:
            if safe:
                return self.tmp_R.copy()
            return self.tmp_R
        return self.gd.distribute(self.tmp_R)

    def scatter(self, a_G, q=None):
        """Scatter coefficients from master to all cores."""
        comm = self.gd.comm
        if comm.size == 1:
            return a_G

        mya_G = np.empty(self.maxmyng, complex)
        comm.scatter(pad(a_G, self.maxmyng * comm.size), mya_G, 0)
        return mya_G[:self.myng_q[q or 0]]

    def gather(self, a_G, q=None):
        """Gather coefficients on master."""
        comm = self.gd.comm

        if comm.size == 1:
            return a_G

        mya_G = pad(a_G, self.maxmyng)
        if comm.rank == 0:
            a_G = self.tmp_G
        else:
            a_G = None
        comm.gather(mya_G, 0, a_G)
        if comm.rank == 0:
            return a_G[:self.ng_q[q or 0]]

    def alltoall1(self, a_rG, q):
        """Gather coefficients from a_rG[r] on rank r.

        On rank r, an array of all G-vector coefficients will be returned.
        These will be gathered from a_rG[r] on all the cores.
        """
        comm = self.gd.comm
        if comm.size == 1:
            return a_rG[0]
        N = len(a_rG)
        ng = self.ng_q[q]
        ssize_r = np.zeros(comm.size, int)
        ssize_r[:N] = self.myng_q[q]
        soffset_r = np.arange(comm.size) * self.myng_q[q]
        soffset_r[N:] = 0
        roffset_r = (np.arange(comm.size) * self.maxmyng).clip(max=ng)
        rsize_r = np.zeros(comm.size, int)
        if comm.rank < N:
            rsize_r[:-1] = roffset_r[1:] - roffset_r[:-1]
            rsize_r[-1] = ng - roffset_r[-1]
        b_G = self.tmp_G[:ng]
        comm.alltoallv(a_rG, ssize_r, soffset_r, b_G, rsize_r, roffset_r)
        if comm.rank < N:
            return b_G

    def alltoall2(self, a_G, q, b_rG):
        """Scatter all coefs. from rank r to B_rG[r] on other cores."""
        comm = self.gd.comm
        if comm.size == 1:
            b_rG[0] += a_G
            return
        N = len(b_rG)
        ng = self.ng_q[q]
        rsize_r = np.zeros(comm.size, int)
        rsize_r[:N] = self.myng_q[q]
        roffset_r = np.arange(comm.size) * self.myng_q[q]
        roffset_r[N:] = 0
        soffset_r = (np.arange(comm.size) * self.maxmyng).clip(max=ng)
        ssize_r = np.zeros(comm.size, int)
        if comm.rank < N:
            ssize_r[:-1] = soffset_r[1:] - soffset_r[:-1]
            ssize_r[-1] = ng - soffset_r[-1]
        tmp_rG = self.tmp_G[:b_rG.size].reshape(b_rG.shape)
        comm.alltoallv(a_G, ssize_r, soffset_r, tmp_rG, rsize_r, roffset_r)
        b_rG += tmp_rG

    def integrate(self, a_xg, b_yg=None,
                  global_integral=True, hermitian=False):
        """Integrate function(s) over domain.

        a_xg: ndarray
            Function(s) to be integrated.
        b_yg: ndarray
            If present, integrate a_xg.conj() * b_yg.
        global_integral: bool
            If the array(s) are distributed over several domains, then the
            total sum will be returned.  To get the local contribution
            only, use global_integral=False.
        hermitian: bool
            Result is hermitian.
        """

        if b_yg is None:
            # Only one array:
            assert self.dtype == float and self.gd.comm.size == 1
            return a_xg[..., 0].real * self.gd.dv

        if a_xg.ndim == 1:
            A_xg = a_xg.reshape((1, len(a_xg)))
        else:
            A_xg = a_xg
        if b_yg.ndim == 1:
            B_yg = b_yg.reshape((1, len(b_yg)))
        else:
            B_yg = b_yg

        alpha = self.gd.dv / self.gd.N_c.prod()

        if self.dtype == float:
            alpha *= 2
            A_xg = A_xg.view(float)
            B_yg = B_yg.view(float)

        result_yx = np.zeros((len(B_yg), len(A_xg)), self.dtype)

        if a_xg is b_yg:
            rk(alpha, A_xg, 0.0, result_yx)
        elif hermitian:
            r2k(0.5 * alpha, A_xg, B_yg, 0.0, result_yx)
        else:
            mmm(alpha, B_yg, 'N', A_xg, 'C', 0.0, result_yx)

        if self.dtype == float and self.gd.comm.rank == 0:
            correction_yx = np.outer(B_yg[:, 0], A_xg[:, 0])
            if hermitian:
                result_yx -= 0.25 * alpha * (correction_yx + correction_yx.T)
            else:
                result_yx -= 0.5 * alpha * correction_yx

        xshape = a_xg.shape[:-1]
        yshape = b_yg.shape[:-1]
        result = result_yx.T.reshape(xshape + yshape)

        if result.ndim == 0:
            if global_integral:
                return self.gd.comm.sum_scalar(result.item())
            return result.item()
        else:
            assert global_integral or self.gd.comm.size == 1
            self.gd.comm.sum(result.T)
            return result

    def interpolate(self, a_R, pd):
        if (pd.gd.N_c <= self.gd.N_c).any():
            raise ValueError('Too few points in target grid!')

        self.gd.collect(a_R, self.tmp_R[:])

        if self.gd.comm.rank == 0:
            self.fftplan.execute()

            a_Q = self.tmp_Q
            b_Q = pd.tmp_Q

            e0, e1, e2 = 1 - self.gd.N_c % 2  # even or odd size
            a0, a1, a2 = pd.gd.N_c // 2 - self.gd.N_c // 2
            b0, b1, b2 = self.gd.N_c + (a0, a1, a2)

            if self.dtype == float:
                b2 = (b2 - a2) // 2 + 1
                a2 = 0
                axes = (0, 1)
            else:
                axes = (0, 1, 2)

            b_Q[:] = 0.0
            b_Q[a0:b0, a1:b1, a2:b2] = np.fft.fftshift(a_Q, axes=axes)

            if e0:
                b_Q[a0, a1:b1, a2:b2] *= 0.5
                b_Q[b0, a1:b1, a2:b2] = b_Q[a0, a1:b1, a2:b2]
                b0 += 1
            if e1:
                b_Q[a0:b0, a1, a2:b2] *= 0.5
                b_Q[a0:b0, b1, a2:b2] = b_Q[a0:b0, a1, a2:b2]
                b1 += 1
            if self.dtype == complex:
                if e2:
                    b_Q[a0:b0, a1:b1, a2] *= 0.5
                    b_Q[a0:b0, a1:b1, b2] = b_Q[a0:b0, a1:b1, a2]
            else:
                if e2:
                    b_Q[a0:b0, a1:b1, b2 - 1] *= 0.5

            b_Q[:] = np.fft.ifftshift(b_Q, axes=axes)
            pd.ifftplan.execute()

            a_G = a_Q.ravel()[self.Q_qG[0]]
        else:
            a_G = None

        return (pd.gd.distribute(pd.tmp_R) * (1.0 / self.tmp_R.size),
                self.scatter(a_G))

    def restrict(self, a_R, pd):
        self.gd.collect(a_R, self.tmp_R[:])

        if self.gd.comm.rank == 0:
            a_Q = pd.tmp_Q
            b_Q = self.tmp_Q

            e0, e1, e2 = 1 - pd.gd.N_c % 2  # even or odd size
            a0, a1, a2 = self.gd.N_c // 2 - pd.gd.N_c // 2
            b0, b1, b2 = pd.gd.N_c // 2 + self.gd.N_c // 2 + 1

            if self.dtype == float:
                b2 = pd.gd.N_c[2] // 2 + 1
                a2 = 0
                axes = (0, 1)
            else:
                axes = (0, 1, 2)

            self.fftplan.execute()
            b_Q[:] = np.fft.fftshift(b_Q, axes=axes)

            if e0:
                b_Q[a0, a1:b1, a2:b2] += b_Q[b0 - 1, a1:b1, a2:b2]
                b_Q[a0, a1:b1, a2:b2] *= 0.5
                b0 -= 1
            if e1:
                b_Q[a0:b0, a1, a2:b2] += b_Q[a0:b0, b1 - 1, a2:b2]
                b_Q[a0:b0, a1, a2:b2] *= 0.5
                b1 -= 1
            if self.dtype == complex and e2:
                b_Q[a0:b0, a1:b1, a2] += b_Q[a0:b0, a1:b1, b2 - 1]
                b_Q[a0:b0, a1:b1, a2] *= 0.5
                b2 -= 1

            a_Q[:] = b_Q[a0:b0, a1:b1, a2:b2]
            a_Q[:] = np.fft.ifftshift(a_Q, axes=axes)
            a_G = a_Q.ravel()[pd.Q_qG[0]] / 8
            pd.ifftplan.execute()
        else:
            a_G = None

        return (pd.gd.distribute(pd.tmp_R) * (1.0 / self.tmp_R.size),
                pd.scatter(a_G))


class PWMapping:
    def __init__(self, pd1, pd2):
        """Mapping from pd1 to pd2."""
        N_c = np.array(pd1.tmp_Q.shape)
        N2_c = pd2.tmp_Q.shape
        Q1_G = pd1.Q_qG[0]
        Q1_Gc = np.empty((len(Q1_G), 3), int)
        Q1_Gc[:, 0], r_G = divmod(Q1_G, N_c[1] * N_c[2])
        Q1_Gc.T[1:] = divmod(r_G, N_c[2])
        if pd1.dtype == float:
            C = 2
        else:
            C = 3
        Q1_Gc[:, :C] += N_c[:C] // 2
        Q1_Gc[:, :C] %= N_c[:C]
        Q1_Gc[:, :C] -= N_c[:C] // 2
        Q1_Gc[:, :C] %= N2_c[:C]
        Q2_G = Q1_Gc[:, 2] + N2_c[2] * (Q1_Gc[:, 1] + N2_c[1] * Q1_Gc[:, 0])
        G2_Q = np.empty(N2_c, int).ravel()
        G2_Q[:] = -1
        G2_Q[pd2.myQ_qG[0]] = np.arange(pd2.myng_q[0])
        G2_G1 = G2_Q[Q2_G]

        mask_G1 = (G2_G1 != -1)
        self.G2_G1 = G2_G1[mask_G1]
        self.G1 = np.arange(pd1.ngmax)[mask_G1]

        self.pd1 = pd1
        self.pd2 = pd2

    def add_to1(self, a_G1, b_G2):
        """Do a += b * scale, where a is on pd1 and b on pd2."""
        scale = self.pd1.tmp_R.size / self.pd2.tmp_R.size

        if self.pd1.gd.comm.size == 1:
            a_G1 += b_G2[self.G2_G1] * scale
            return

        b_G1 = self.pd1.tmp_G
        b_G1[:] = 0.0
        b_G1[self.G1] = b_G2[self.G2_G1]
        self.pd1.gd.comm.sum(b_G1)
        ng1 = self.pd1.gd.comm.rank * self.pd1.maxmyng
        ng2 = ng1 + self.pd1.myng_q[0]
        a_G1 += b_G1[ng1:ng2] * scale

    def add_to2(self, a_G2, b_G1):
        """Do a += b * scale, where a is on pd2 and b on pd1."""
        myb_G1 = b_G1 * (self.pd2.tmp_R.size / self.pd1.tmp_R.size)
        if self.pd1.gd.comm.size == 1:
            a_G2[self.G2_G1] += myb_G1
            return

        b_G1 = self.pd1.tmp_G
        self.pd1.gd.comm.all_gather(pad(myb_G1, self.pd1.maxmyng), b_G1)
        a_G2[self.G2_G1] += b_G1[self.G1]


def count_reciprocal_vectors(ecut, gd, q_c):
    assert gd.comm.size == 1
    N_c = gd.N_c
    i_Qc = np.indices(N_c).transpose((1, 2, 3, 0))
    i_Qc += N_c // 2
    i_Qc %= N_c
    i_Qc -= N_c // 2

    B_cv = 2.0 * pi * gd.icell_cv
    i_Qc.shape = (-1, 3)
    Gpq_Qv = np.dot(i_Qc, B_cv) + np.dot(q_c, B_cv)

    G2_Q = (Gpq_Qv**2).sum(axis=1)
    return (G2_Q <= 2 * ecut).sum()


def pad(array, N):
    """Pad 1-d ndarray with zeros up to length N.

    >>> a = np.ones(2, complex)
    >>> pad(a, 3)
    array([1.+0.j, 1.+0.j, 0.+0.j])
    >>> pad(a, 2) is a
    True
    >>> pad(None, 7) is None
    True
    """
    if array is None:
        return None
    n = len(array)
    if n == N:
        return array
    if isinstance(array, np.ndarray):
        b = np.empty(N, complex)
    else:
        b = cp.empty(N, complex)
    b[:n] = array
    b[n:] = 0
    return b



import pytest
from gpaw.setup import create_setup


def test_sg15_ni():
    try:
        s = create_setup('Ni', 'PBE', type='sg15')
    except FileNotFoundError:
        pytest.skip(msg='No SG15 PP for Ni found')
    f_sJ = s.calculate_initial_occupation_numbers(2.0, False, 0.0, 2)
    print(f_sJ)
    up, dn = f_sJ.sum(axis=1)
    assert up - dn == pytest.approx(2.0)



import ase.io
import numpy as np
import pytest
from ase.build import molecule
from ase.calculators.vdwcorrection import vdWTkatchenko09prl
from ase.parallel import barrier, parprint

from gpaw import GPAW
from gpaw.analyse.hirshfeld import HirshfeldPartitioning
from gpaw.analyse.vdwradii import vdWradii
from gpaw.utilities.adjust_cell import adjust_cell


@pytest.mark.later
def test_vdw_ts09(in_tmp_dir):
    h = 0.4
    s = molecule('LiH')
    adjust_cell(s, 3., h=h)

    def print_charge_and_check(hp, q=0, label='unpolarized'):
        q_a = np.array(hp.get_charges())
        parprint('Charges ({0})='.format(label), q_a, ', sum=', q_a.sum())
        assert q_a.sum() == pytest.approx(q, abs=0.03)
        return q_a

    # spin unpolarized

    if 1:
        out_traj = 'LiH.traj'
        out_txt = 'LiH.txt'

        cc = GPAW(mode='fd', h=h, xc='PBE', txt=out_txt)

        # this is needed to initialize txt output
        cc.initialize(s)

        hp = HirshfeldPartitioning(cc)
        c = vdWTkatchenko09prl(hp,
                               vdWradii(s.get_chemical_symbols(), 'PBE'))
        s.calc = c
        E = s.get_potential_energy()
        F_ac = s.get_forces()
        s.write(out_traj)
        q_a = print_charge_and_check(hp)

        barrier()

        # test I/O, accuracy due to text output
        accuracy = 1.e-5
        for fname in [out_traj, out_txt]:
            s_out = ase.io.read(fname)
            assert s_out.get_potential_energy() == pytest.approx(E,
                                                                 abs=accuracy)
            for fi, fo in zip(F_ac, s_out.get_forces()):
                assert fi == pytest.approx(fo, abs=accuracy)

    # spin polarized

    if 0:
        ccs = GPAW(mode='fd', h=h, xc='PBE', spinpol=True,
                   txt=None)
        hps = HirshfeldPartitioning(ccs)
        cs = vdWTkatchenko09prl(hps, vdWradii(s.get_chemical_symbols(), 'PBE'))
        s.calc = cs
        Es = s.get_potential_energy()
        Fs_ac = s.get_forces()

        qs_a = print_charge_and_check(hps, label='spin')

        assert q_a == pytest.approx(qs_a, abs=1.e-6)
        assert E == pytest.approx(Es, abs=1.e-4)
        assert F_ac == pytest.approx(Fs_ac, abs=1.e-4)

    # charged

    if 0:
        cc = cc.new(charge=1)
        hpp = HirshfeldPartitioning(cc)
        cp = vdWTkatchenko09prl(hpp,
                                vdWradii(s.get_chemical_symbols(), 'PBE'))
        s.calc = cp
        E = s.get_potential_energy()

        print_charge_and_check(hpp, 1, label='+1')



from __future__ import annotations

from functools import cached_property
from typing import Generator, Generic, TypeVar, TYPE_CHECKING

import numpy as np
from ase.io.ulm import Writer
from ase.units import Bohr, Ha
from gpaw.gpu import as_np, synchronize
from gpaw.gpu.mpi import CuPyMPI
from gpaw.mpi import MPIComm, serial_comm
from gpaw.new import zips
from gpaw.new.brillouin import IBZ
from gpaw.new.c import GPU_AWARE_MPI
from gpaw.new.potential import Potential
from gpaw.new.pwfd.wave_functions import PWFDWaveFunctions
from gpaw.new.wave_functions import WaveFunctions
from gpaw.typing import Array1D, Array2D, Self

if TYPE_CHECKING:
    from gpaw.new.density import Density

WFT = TypeVar('WFT', bound=WaveFunctions)


class IBZWaveFunctions(Generic[WFT]):
    def __init__(self,
                 ibz: IBZ,
                 *,
                 nelectrons: float,
                 ncomponents: int,
                 wfs_qs: list[list[WFT]],
                 kpt_comm: MPIComm = serial_comm,
                 kpt_band_comm: MPIComm = serial_comm,
                 comm: MPIComm = serial_comm):
        """Collection of wave function objects for k-points in the IBZ."""
        self.ibz = ibz
        self.kpt_comm = kpt_comm
        self.kpt_band_comm = kpt_band_comm
        self.comm = comm
        self.nelectrons = nelectrons
        self.ncomponents = ncomponents
        self.collinear = (ncomponents != 4)
        self.spin_degeneracy = ncomponents % 2 + 1
        self.nspins = ncomponents % 3

        self.rank_k = ibz.ranks(kpt_comm)

        self.wfs_qs = wfs_qs

        self.q_k = {}  # IBZ-index to local index
        for wfs in self:
            self.q_k[wfs.k] = wfs.q

        self.band_comm = wfs.band_comm
        self.domain_comm = wfs.domain_comm
        self.dtype = wfs.dtype
        self.nbands = wfs.nbands

        self.fermi_levels: Array1D | None = None  # hartree

        self.energies: dict[str, float] = {}  # hartree

        self.xp = self.wfs_qs[0][0].xp
        if self.xp is not np:
            if not GPU_AWARE_MPI:
                self.kpt_comm = CuPyMPI(self.kpt_comm)  # type: ignore

    @classmethod
    def create(cls,
               *,
               ibz: IBZ,
               nelectrons: float,
               ncomponents: int,
               create_wfs_func,
               kpt_comm: MPIComm = serial_comm,
               kpt_band_comm: MPIComm = serial_comm,
               comm: MPIComm = serial_comm,
               ) -> Self:
        """Collection of wave function objects for k-points in the IBZ."""
        rank_k = ibz.ranks(kpt_comm)
        mask_k = (rank_k == kpt_comm.rank)
        k_q = np.arange(len(ibz))[mask_k]

        nspins = ncomponents % 3

        wfs_qs: list[list[WFT]] = []
        for q, k in enumerate(k_q):
            wfs_s = []
            for spin in range(nspins):
                wfs = create_wfs_func(spin, q, k,
                                      ibz.kpt_kc[k], ibz.weight_k[k])
                wfs_s.append(wfs)
            wfs_qs.append(wfs_s)

        return cls(ibz,
                   nelectrons=nelectrons,
                   ncomponents=ncomponents,
                   wfs_qs=wfs_qs,
                   kpt_comm=kpt_comm,
                   kpt_band_comm=kpt_band_comm,
                   comm=comm)

    @cached_property
    def mode(self):
        wfs = self.wfs_qs[0][0]
        if isinstance(wfs, PWFDWaveFunctions):
            if hasattr(wfs.psit_nX.desc, 'ecut'):
                return 'pw'
            return 'fd'
        return 'lcao'

    def get_max_shape(self, global_shape: bool = False) -> tuple[int, ...]:
        """Find the largest wave function array shape.

        For a PW-calculation, this shape could depend on k-point.
        """
        if global_shape:
            shape = np.array(max(wfs.array_shape(global_shape=True)
                                 for wfs in self))
            self.kpt_comm.max(shape)
            return tuple(shape)
        return max(wfs.array_shape() for wfs in self)

    @property
    def fermi_level(self) -> float:
        fl = self.fermi_levels
        assert fl is not None and len(fl) == 1
        return fl[0]

    def __str__(self):
        shape = self.get_max_shape(global_shape=True)
        wfs = self.wfs_qs[0][0]
        nbytes = (len(self.ibz) *
                  self.nbands *
                  len(self.wfs_qs[0]) *
                  wfs.bytes_per_band)
        ncores = (self.kpt_comm.size *
                  self.domain_comm.size *
                  self.band_comm.size)
        return (f'{self.ibz.symmetries}\n'
                f'{self.ibz}\n'
                f'{wfs._short_string(shape)}\n'
                f'spin-components: {self.ncomponents}'
                '  # (' +
                ('' if self.collinear else 'non-') + 'collinear spins)\n'
                f'bands: {self.nbands}\n'
                f'valence electrons: {self.nelectrons}\n'
                f'spin-degeneracy: {self.spin_degeneracy}\n'
                f'dtype: {self.dtype}\n\n'
                'memory:\n'
                f'    wave functions: {nbytes:_}  # bytes '
                f' ({nbytes // ncores:_} per core)\n\n'
                'parallelization:\n'
                f'    kpt:    {self.kpt_comm.size}\n'
                f'    domain: {self.domain_comm.size}\n'
                f'    band:   {self.band_comm.size}\n')

    def __iter__(self) -> Generator[WFT, None, None]:
        for wfs_s in self.wfs_qs:
            yield from wfs_s

    def move(self, fracpos_ac, atomdist):
        self.ibz.symmetries.check_positions(fracpos_ac)
        self.energies.clear()
        for wfs in self:
            wfs.move(fracpos_ac, atomdist)

    def orthonormalize(self, work_array_nX: np.ndarray = None):
        for wfs in self:
            wfs.orthonormalize(work_array_nX)

    def calculate_occs(self, occ_calc, fixed_fermi_level=False):
        degeneracy = self.spin_degeneracy

        # u index is q and s combined
        occ_un, fermi_levels, e_entropy = occ_calc.calculate(
            nelectrons=self.nelectrons / degeneracy,
            eigenvalues=[wfs.eig_n * Ha for wfs in self],
            weights=[wfs.weight for wfs in self],
            fermi_levels_guess=(None
                                if self.fermi_levels is None else
                                self.fermi_levels * Ha))

        if not fixed_fermi_level:
            self.fermi_levels = np.array(fermi_levels) / Ha
        else:
            assert self.fermi_levels is not None

        for occ_n, wfs in zips(occ_un, self):
            wfs._occ_n = occ_n

        e_entropy *= degeneracy / Ha
        e_band = 0.0
        for wfs in self:
            e_band += wfs.occ_n @ wfs.eig_n * wfs.weight * degeneracy
        e_band = self.kpt_comm.sum_scalar(float(e_band))  # XXX CPU float?

        self.energies.update(
            band=e_band,
            entropy=e_entropy,
            extrapolation=e_entropy * occ_calc.extrapolate_factor)

    def add_to_density(self, nt_sR, D_asii) -> None:
        """Compute density and add to ``nt_sR`` and ``D_asii``."""
        for wfs in self:
            wfs.add_to_density(nt_sR, D_asii)

        if self.xp is not np:
            synchronize()

        # This should be done in a more efficient way!!!
        # Also: where do we want the density?
        self.kpt_comm.sum(nt_sR.data)
        self.kpt_comm.sum(D_asii.data)
        self.band_comm.sum(nt_sR.data)
        self.band_comm.sum(D_asii.data)

    def normalize_density(self, density: Density) -> None:
        pass  # overwritten in LCAOIBZWaveFunctions class

    def add_to_ked(self, taut_sR) -> None:
        for wfs in self:
            wfs.add_to_ked(taut_sR)
        if self.xp is not np:
            synchronize()
        self.kpt_comm.sum(taut_sR.data)
        self.band_comm.sum(taut_sR.data)

    def get_all_electron_wave_function(self,
                                       band,
                                       kpt=0,
                                       spin=0,
                                       grid_spacing=0.05,
                                       skip_paw_correction=False):
        wfs = self.get_wfs(kpt=kpt, spin=spin, n1=band, n2=band + 1)
        if wfs is None:
            return None
        assert isinstance(wfs, PWFDWaveFunctions)
        psit_X = wfs.psit_nX[0].to_pbc_grid()
        grid = psit_X.desc.uniform_grid_with_grid_spacing(grid_spacing)
        psi_r = psit_X.interpolate(grid=grid)

        if not skip_paw_correction:
            dphi_aj = wfs.setups.partial_wave_corrections()
            dphi_air = grid.atom_centered_functions(dphi_aj, wfs.fracpos_ac)
            dphi_air.add_to(psi_r, wfs.P_ani[:, 0])

        return psi_r

    def get_wfs(self,
                *,
                kpt: int = 0,
                spin: int = 0,
                n1=0,
                n2=0):
        rank = self.rank_k[kpt]
        if rank == self.kpt_comm.rank:
            wfs = self.wfs_qs[self.q_k[kpt]][spin]
            wfs2 = wfs.collect(n1, n2)
            if rank == 0:
                return wfs2
            if wfs2 is not None:
                wfs2.send(0, self.kpt_comm)
            return
        if self.comm.rank == 0:
            return self.wfs_qs[0][0].receive(rank, self.kpt_comm)
        return None

    def get_eigs_and_occs(self, k=0, s=0):
        if self.domain_comm.rank == 0 and self.band_comm.rank == 0:
            rank = self.rank_k[k]
            if rank == self.kpt_comm.rank:
                wfs = self.wfs_qs[self.q_k[k]][s]
                if rank == 0:
                    return wfs._eig_n, wfs._occ_n
                self.kpt_comm.send(wfs._eig_n, 0)
                self.kpt_comm.send(wfs._occ_n, 0)
            elif self.kpt_comm.rank == 0:
                eig_n = np.empty(self.nbands)
                occ_n = np.empty(self.nbands)
                self.kpt_comm.receive(eig_n, rank)
                self.kpt_comm.receive(occ_n, rank)
                return eig_n, occ_n
        return np.zeros(0), np.zeros(0)

    def get_all_eigs_and_occs(self):
        nkpts = len(self.ibz)
        if self.comm.rank == 0:
            eig_skn = np.empty((self.nspins, nkpts, self.nbands))
            occ_skn = np.empty((self.nspins, nkpts, self.nbands))
        else:
            eig_skn = np.empty((0, 0, 0))
            occ_skn = np.empty((0, 0, 0))
        for k in range(nkpts):
            for s in range(self.nspins):
                eig_n, occ_n = self.get_eigs_and_occs(k, s)
                if self.comm.rank == 0:
                    eig_skn[s, k, :] = eig_n
                    occ_skn[s, k, :] = occ_n
        return eig_skn, occ_skn

    def forces(self, potential: Potential) -> Array2D:
        self.make_sure_wfs_are_read_from_gpw_file()
        F_av = self.xp.zeros((potential.dH_asii.natoms, 3))
        for wfs in self:
            wfs.force_contribution(potential, F_av)
        if self.xp is not np:
            synchronize()
        self.kpt_band_comm.sum(F_av)
        return F_av

    def write(self,
              writer: Writer,
              skip_wfs: bool) -> None:
        """Write fermi-level(s), eigenvalues, occupation numbers, ...

        ... k-points, symmetry information, projections and possibly
        also the wave functions.
        """
        eig_skn, occ_skn = self.get_all_eigs_and_occs()
        assert self.fermi_levels is not None
        writer.write(fermi_levels=self.fermi_levels * Ha,
                     eigenvalues=eig_skn * Ha,
                     occupations=occ_skn)
        ibz = self.ibz
        writer.child('kpts').write(
            atommap=ibz.symmetries.a_sa,
            bz2ibz=ibz.bz2ibz_K,
            bzkpts=ibz.bz.kpt_Kc,
            ibzkpts=ibz.kpt_kc,
            rotations=ibz.symmetries.rotation_svv,
            translations=ibz.symmetries.translation_sc,
            weights=ibz.weight_k)

        nproj = self.wfs_qs[0][0].P_ani.layout.size

        spin_k_shape: tuple[int, ...]
        proj_shape: tuple[int, ...]

        if self.collinear:
            spin_k_shape = (self.ncomponents, len(ibz))
            proj_shape = (self.nbands, nproj)
        else:
            spin_k_shape = (len(ibz),)
            proj_shape = (self.nbands, 2, nproj)

        writer.add_array('projections', spin_k_shape + proj_shape, self.dtype)

        for spin in range(self.nspins):
            for k, rank in enumerate(self.rank_k):
                if rank == self.kpt_comm.rank:
                    wfs = self.wfs_qs[self.q_k[k]][spin]
                    P_ani = wfs.P_ani.to_cpu().gather()  # gather atoms
                    if P_ani is not None:
                        P_nI = P_ani.matrix.gather()  # gather bands
                        if P_nI.dist.comm.rank == 0:
                            if rank == 0:
                                writer.fill(P_nI.data.reshape(proj_shape))
                            else:
                                self.kpt_comm.send(P_nI.data, 0)
                elif self.comm.rank == 0:
                    data = np.empty(proj_shape, self.dtype)
                    self.kpt_comm.receive(data, rank)
                    writer.fill(data)

        if not skip_wfs:
            self._write_wave_functions(writer, spin_k_shape)

    def _write_wave_functions(self, writer, spin_k_shape):
        # We collect all bands to master.  This may have to be changed
        # to only one band at a time XXX
        xshape = self.get_max_shape(global_shape=True)
        shape = spin_k_shape + (self.nbands,) + xshape
        dtype = complex if self.mode == 'pw' else self.dtype
        c = 1.0 if self.mode == 'lcao' else Bohr**-1.5

        writer.add_array('coefficients', shape, dtype=dtype)
        buf_nX = np.empty((self.nbands,) + xshape, dtype=dtype)

        for spin in range(self.nspins):
            for k, rank in enumerate(self.rank_k):
                if rank == self.kpt_comm.rank:
                    wfs = self.wfs_qs[self.q_k[k]][spin]
                    coef_nX = wfs.gather_wave_function_coefficients()
                    if coef_nX is not None:
                        coef_nX = as_np(coef_nX)
                        if self.mode == 'pw':
                            x = coef_nX.shape[-1]
                            if x < xshape[-1]:
                                # For PW-mode, we may need to zero-pad the
                                # plane-wave coefficient up to the maximum
                                # for all k-points:
                                buf_nX[..., :x] = coef_nX
                                buf_nX[..., x:] = 0.0
                                coef_nX = buf_nX
                        if rank == 0:
                            writer.fill(coef_nX * c)
                        else:
                            self.kpt_comm.send(coef_nX, 0)
                elif self.comm.rank == 0:
                    self.kpt_comm.receive(buf_nX, rank)
                    writer.fill(buf_nX * c)

    def write_summary(self, log):
        fl = self.fermi_levels * Ha
        if len(fl) == 1:
            log(f'\nFermi level: {fl[0]:.3f}')
        else:
            log(f'\nFermi levels: {fl[0]:.3f}, {fl[1]:.3f}')

        ibz = self.ibz

        eig_skn, occ_skn = self.get_all_eigs_and_occs()

        if self.comm.rank != 0:
            return

        eig_skn *= Ha

        D = self.spin_degeneracy
        nbands = eig_skn.shape[2]

        for k, (x, y, z) in enumerate(ibz.kpt_kc):
            if k == 3:
                log(f'(only showing first 3 out of {len(ibz)} k-points)')
                break

            log(f'\nkpt = [{x:.3f}, {y:.3f}, {z:.3f}], '
                f'weight = {ibz.weight_k[k]:.3f}:')

            if self.nspins == 1:
                skipping = False
                log(f'  Band      eig [eV]   occ [0-{D}]')
                eig_n = eig_skn[0, k]
                n0 = (eig_n < fl[0]).sum() - 0.5
                for n, (e, f) in enumerate(zips(eig_n, occ_skn[0, k])):
                    # First, last and +-8 bands window around Fermi level:
                    if n == 0 or abs(n - n0) < 8 or n == nbands - 1:
                        log(f'  {n:4} {e:13.3f}   {D * f:9.3f}')
                        skipping = False
                    else:
                        if not skipping:
                            log('   ...')
                            skipping = True
            else:
                log('  Band      eig [eV]   occ [0-1]'
                    '      eig [eV]   occ [0-1]')
                for n, (e1, f1, e2, f2) in enumerate(zips(eig_skn[0, k],
                                                          occ_skn[0, k],
                                                          eig_skn[1, k],
                                                          occ_skn[1, k])):
                    log(f'  {n:4} {e1:13.3f}   {f1:9.3f}'
                        f'    {e2:10.3f}   {f2:9.3f}')

        try:
            from ase.dft.bandgap import GapInfo
        except ImportError:
            log('No gapinfo -- requires new ASE')
            return

        try:
            log()
            fermilevel = fl[0]
            gapinfo = GapInfo(eigenvalues=eig_skn - fermilevel)
            log(gapinfo.description(ibz_kpoints=ibz.kpt_kc))
        except ValueError:
            # Maybe we only have the occupied bands and no empty bands
            log('Could not find a gap')

    def make_sure_wfs_are_read_from_gpw_file(self):
        for wfs in self:
            psit_nX = getattr(wfs, 'psit_nX', None)
            if psit_nX is None:
                return
            if hasattr(psit_nX.data, 'fd'):  # fd=file-descriptor
                psit_nX.data = psit_nX.data[:]  # read

    def get_homo_lumo(self, spin: int = None) -> Array1D:
        """Return HOMO and LUMO eigenvalues."""
        if self.ncomponents == 1:
            N = 2
            assert spin != 1
            spin = 0
        elif self.ncomponents == 2:
            N = 2
            if spin is None:
                h0, l0 = self.get_homo_lumo(0)
                h1, l1 = self.get_homo_lumo(1)
                return np.array([max(h0, h1), min(l0, l1)])
        else:
            N = 1
            assert spin != 1
            spin = 0

        n = int(round(self.nelectrons)) // N
        assert N * n == self.nelectrons
        homo = self.kpt_comm.max_scalar(max(wfs_s[spin].eig_n[n - 1]
                                            for wfs_s in self.wfs_qs))
        lumo = self.kpt_comm.min_scalar(min(wfs_s[spin].eig_n[n]
                                            for wfs_s in self.wfs_qs))

        return np.array([homo, lumo])



import numpy as np
import pytest

from ase.units import Bohr, Hartree
from gpaw.external import ConstantElectricField
from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.lcaotddft.laser import create_laser, register_custom_laser, Laser
from gpaw.mpi import world
from gpaw.tddft.units import as_to_au

# Settings
dt = 20.0
N1 = 5
N = 5 + N1
kick_v = np.ones(3) * 1e-5


@pytest.mark.rttddft
@pytest.mark.parametrize('pulse', [
    {'name': 'GaussianPulse', 'strength': 1e-5, 'time0': 0, 'frequency': 8.6,
     'sigma': 0.5, 'sincos': 'sin'},
    {'name': 'SincPulse', 'strength': 1e-6, 'time0': 0, 'cutoff_freq': 15,
     'relative_t0': False}])
def test_laser(gpw_files, in_tmp_dir, pulse):
    # Time-propagation calculation
    td_calc = LCAOTDDFT(gpw_files['na2_tddft_dzp'], txt='td.out')
    DipoleMomentWriter(td_calc, 'dm.dat')
    td_calc.absorption_kick(kick_v)
    td_calc.propagate(dt, N)

    # Pulse
    direction = kick_v
    ext = ConstantElectricField(Hartree / Bohr, direction)

    # Time-propagation calculation with pulse
    pulse = create_laser(pulse)
    td_calc = LCAOTDDFT(gpw_files['na2_tddft_dzp'],
                        td_potential={'ext': ext, 'laser': pulse},
                        txt='tdpulse.out')
    DipoleMomentWriter(td_calc, 'dmpulse.dat')
    td_calc.propagate(dt, N1)
    td_calc.write('td.gpw', mode='all')
    # Restart
    td_calc = LCAOTDDFT('td.gpw', txt='tdpulse2.out')
    DipoleMomentWriter(td_calc, 'dmpulse.dat')
    td_calc.propagate(dt, N - N1)

    # Convoluted dipole moment
    world.barrier()
    time_t = np.arange(0, dt * (N + 0.1), dt) * as_to_au
    pulse_t = pulse.strength(time_t)
    np.savetxt('pulse.dat', np.stack((time_t, pulse_t)).T)
    dm_tv = np.delete(np.loadtxt('dm.dat')[:, 2:], 1, axis=0)
    dm_tv /= np.linalg.norm(kick_v)
    pulsedm_tv = np.delete(np.loadtxt('dmpulse.dat')[:, 2:], N1, axis=0)

    tol = 5e-6
    for v in range(3):
        pulsedmconv_t = np.convolve(
            dm_tv[:, v], pulse_t)[:(N + 1)] * dt * as_to_au
        np.savetxt('dmpulseconv%d.dat' % v, pulsedmconv_t)
        assert pulsedm_tv[:, v] == pytest.approx(pulsedmconv_t, abs=tol)


@pytest.mark.rttddft
def test_custom(gpw_files):
    gpw_fname = gpw_files['na2_tddft_dzp']

    class RandomPulse(Laser):
        def __init__(self, strength):
            self.rng = np.random.default_rng(42)
            self.dict = dict(name='RandomPulse', strength=strength)
            self.s0 = strength

        def strength(self, t):
            return self.s0 * self.rng.uniform(size=np.shape(t))

        def todict(self):
            return self.dict

    # We should be able to run and restart using custom pulses
    register_custom_laser('RandomPulse', RandomPulse)

    # Pulse
    direction = kick_v
    ext = ConstantElectricField(Hartree / Bohr, direction)

    # Time-propagation calculation with pulse
    pulse = RandomPulse(1e-5)
    td_calc = LCAOTDDFT(gpw_fname, td_potential={'ext': ext, 'laser': pulse},
                        txt='tdpulse.out')
    td_calc.propagate(dt, 1)

    td_calc.write('td.gpw', mode='all')

    # Restart
    td_calc = LCAOTDDFT('td.gpw', txt='tdpulse2.out')

    restart_pulse = td_calc.td_hamiltonian.td_potential.laser_i[0]

    assert isinstance(restart_pulse, RandomPulse)
    np.testing.assert_equal(restart_pulse.todict(), pulse.todict())



""" Maximally localized Wannier Functions

    Find the set of maximally localized Wannier functions
    using the spread functional of Marzari and Vanderbilt
    (PRB 56, 1997 page 12847).

    this code is as in ASE but modified to use it with gpaw's wfs.
"""

from time import time
from math import pi
import numpy as np
from ase.dft.kpoints import get_monkhorst_pack_size_and_offset
from ase.dft.wannier import calculate_weights, gram_schmidt
from ase.transport.tools import dagger
from ase.parallel import parprint

dag = dagger


def random_orthogonal_matrix(dim, rng, real=False):
    """Generate a random orthogonal matrix"""

    H = rng.random((dim, dim))
    np.add(dag(H), H, H)
    np.multiply(.5, H, H)

    if real:
        gram_schmidt(H)
        return H
    else:
        val, vec = np.linalg.eig(H)
        return np.dot(vec * np.exp(1.j * val), dag(vec))


def md_min(func, step=.25, tolerance=1e-6, verbose=False, **kwargs):
    if verbose:
        parprint('Localize with step =', step,
                 'and tolerance =', tolerance)
    t = -time()
    fvalueold = 0.
    fvalue = fvalueold + 10
    count = 0
    V = np.zeros(func.get_gradients().shape, dtype=complex)

    while abs((fvalue - fvalueold) / fvalue) > tolerance:
        fvalueold = fvalue
        dF = func.get_gradients()
        V *= (dF * V.conj()).real > 0
        V += step * dF
        func.step(V, **kwargs)
        fvalue = func.get_function_value()

        if fvalue < fvalueold:
            step *= 0.5
        count += 1
        func.niter = count

        if verbose:
            parprint('MDmin: iter=%s, step=%s, value=%s'
                     % (count, step, fvalue))
    t += time()
    if verbose:
        parprint('%d iterations in %0.2f seconds(%0.2f ms/iter),'
                 ' endstep = %s'
                 % (count, t, t * 1000. / count, step))


def get_atoms_object_from_wfs(wfs):
    from ase.units import Bohr
    from ase import Atoms

    spos_ac = wfs.spos_ac
    cell_cv = wfs.gd.cell_cv
    positions = spos_ac * cell_cv.diagonal() * Bohr

    string = ''
    for a, atoms in enumerate(wfs.setups):
        string += atoms.symbol

    atoms = Atoms(string)
    atoms.positions = positions
    atoms.cell = cell_cv * Bohr

    return atoms


class WannierLocalization:
    """Maximally localized Wannier Functions
       for n_occ only - for ODD calculations
    """

    def __init__(self, wfs, calc=None, spin=0, seed=None, verbose=False):
        from ase.dft.wannier import get_kklst, get_invkklst

        # Bloch phase sign convention
        sign = -1
        self.wfs = wfs
        self.gd = self.wfs.gd
        self.ns = self.wfs.nspins
        self.dtype = wfs.dtype

        if hasattr(self.wfs, 'mode'):
            self.mode = self.wfs.mode
        else:
            self.mode = None

        if calc is not None:
            self.atoms = calc.atoms
        else:
            self.atoms = get_atoms_object_from_wfs(self.wfs)

        # Determine nocc: integer occupations only
        k_rank, u = divmod(0 + len(self.wfs.kd.ibzk_kc) * spin,
                           len(self.wfs.kpt_u))

        f_n = self.wfs.kpt_u[u].f_n
        self.nwannier = int(np.rint(f_n.sum()) /
                            (3 - self.ns))  # No fractional occ

        self.spin = spin
        self.verbose = verbose
        self.rng = np.random.default_rng(seed)
        self.kpt_kc = self.wfs.kd.bzk_kc
        assert len(self.wfs.kd.ibzk_kc) == len(self.kpt_kc)

        self.kptgrid = \
            get_monkhorst_pack_size_and_offset(self.kpt_kc)[0]
        self.kpt_kc *= sign

        self.Nk = len(self.kpt_kc)
        self.unitcell_cc = self.atoms.get_cell()
        self.largeunitcell_cc = (self.unitcell_cc.T * self.kptgrid).T
        self.weight_d, self.Gdir_dc = \
            calculate_weights(self.largeunitcell_cc)
        self.Ndir = len(self.weight_d)  # Number of directions

        # Get neighbor kpt list and inverse kpt list
        self.kklst_dk, k0_dkc = get_kklst(self.kpt_kc, self.Gdir_dc)
        self.invkklst_dk = get_invkklst(self.kklst_dk)

        Nw = self.nwannier
        Z_dknn = np.zeros((self.Ndir, self.Nk, Nw, Nw),
                          dtype=complex)
        self.Z_dkww = np.empty((self.Ndir, self.Nk, Nw, Nw),
                               dtype=complex)

        if self.mode == 'lcao' and self.wfs.kpt_u[0].psit_nG is None:
            self.wfs.initialize_wave_functions_from_lcao()

        for d, dirG in enumerate(self.Gdir_dc):
            for k in range(self.Nk):
                k1 = self.kklst_dk[d, k]
                k0_c = k0_dkc[d, k]
                k_kc = self.wfs.kd.bzk_kc
                Gc = k_kc[k1] - k_kc[k] - k0_c
                # Det. kpt/spin
                kr, u = divmod(k + len(self.wfs.kd.ibzk_kc) * spin,
                               len(self.wfs.kpt_u))
                kr1, u1 = divmod(k1 + len(self.wfs.kd.ibzk_kc) * spin,
                                 len(self.wfs.kpt_u))

                if self.wfs.mode == 'pw':
                    cmo = self.gd.zeros(Nw, dtype=self.wfs.dtype)
                    cmo1 = self.gd.zeros(Nw, dtype=self.wfs.dtype)
                    for i in range(Nw):
                        cmo[i] = self.wfs._get_wave_function_array(u, i)
                        cmo1[i] = self.wfs._get_wave_function_array(u1, i)
                else:
                    cmo = self.wfs.kpt_u[u].psit_nG[:Nw]
                    cmo1 = self.wfs.kpt_u[u1].psit_nG[:Nw]

                e_G = np.exp(-2.j * pi *
                             np.dot(np.indices(self.gd.n_c).T +
                                    self.gd.beg_c,
                                    Gc / self.gd.N_c).T)
                pw = (e_G * cmo.conj()).reshape((Nw, -1))

                Z_dknn[d, k] += \
                    np.inner(pw, cmo1.reshape((Nw, -1))) * self.gd.dv
                # PAW corrections
                P_ani1 = self.wfs.kpt_u[u1].P_ani
                spos_ac = self.atoms.get_scaled_positions()

                for A, P_ni in self.wfs.kpt_u[u].P_ani.items():
                    dS_ii = self.wfs.setups[A].dO_ii
                    P_n = P_ni[:Nw]
                    P_n1 = P_ani1[A][:Nw]
                    e = np.exp(-2.j * pi * np.dot(Gc, spos_ac[A]))

                    Z_dknn[d, k] += e * P_n.conj().dot(
                        dS_ii.dot(P_n1.T))

        self.gd.comm.sum(Z_dknn)
        self.Z_dknn = Z_dknn.copy()

        self.initialize()

    def initialize(self):
        """Re-initialize current rotation matrix.

        Keywords are identical to those of the constructor.
        """
        Nw = self.nwannier

        # Set U to random (orthogonal) matrix
        self.U_kww = np.zeros((self.Nk, Nw, Nw), self.dtype)

        # for k in range(self.Nk):
        if self.dtype == float:
            real = True
        else:
            real = False
        self.U_kww[:] = random_orthogonal_matrix(Nw, self.rng, real=real)

        self.update()

    def update(self):

        # Calculate the Zk matrix from the rotation matrix:
        # Zk = U^d[k] Zbloch U[k1]
        for d in range(self.Ndir):
            for k in range(self.Nk):
                k1 = self.kklst_dk[d, k]
                self.Z_dkww[d, k] = np.dot(dag(self.U_kww[k]), np.dot(
                    self.Z_dknn[d, k], self.U_kww[k1]))

        # Update the new Z matrix
        self.Z_dww = self.Z_dkww.sum(axis=1) / self.Nk

    def get_centers(self, scaled=False):
        """Calculate the Wannier centers

        ::

          pos =  L / 2pi * phase(diag(Z))
        """
        coord_wc = \
            np.angle(self.Z_dww[:3].diagonal(0, 1, 2)).T / \
            (2.0 * pi) % 1
        if not scaled:
            coord_wc = np.dot(coord_wc, self.largeunitcell_cc)
        return coord_wc

    def localize(self, step=0.25, tolerance=1e-08,
                 updaterot=True):
        """Optimize rotation to give maximal localization"""
        md_min(self, step, tolerance, verbose=self.verbose,
               updaterot=updaterot)

    def get_function_value(self):
        """Calculate the value of the spread functional.

        ::

          Tr[|ZI|^2]=sum(I)sum(n) w_i|Z_(i)_nn|^2,

        where w_i are weights."""
        a_d = np.sum(np.abs(self.Z_dww.diagonal(0, 1, 2)) ** 2,
                     axis=1)
        return np.dot(a_d, self.weight_d).real

    def get_gradients(self):

        Nw = self.nwannier
        dU = []
        for k in range(self.Nk):
            Utemp_ww = np.zeros((Nw, Nw), complex)

            for d, weight in enumerate(self.weight_d):
                if abs(weight) < 1.0e-6:
                    continue

                diagZ_w = self.Z_dww[d].diagonal()
                Zii_ww = np.repeat(diagZ_w, Nw).reshape(Nw, Nw)
                k2 = self.invkklst_dk[d, k]
                Z_kww = self.Z_dkww[d]

                temp = Zii_ww.T * Z_kww[k].conj() - \
                    Zii_ww * Z_kww[k2].conj()
                Utemp_ww += weight * (temp - dag(temp))
            dU.append(Utemp_ww.ravel())

        return np.concatenate(dU)

    def step(self, dX, updaterot=True):
        Nw = self.nwannier
        Nk = self.Nk
        if updaterot:
            A_kww = dX[:Nk * Nw ** 2].reshape(Nk, Nw, Nw)
            for U, A in zip(self.U_kww, A_kww):
                H = -1.j * A.conj()
                epsilon, Z = np.linalg.eigh(H)
                # Z contains the eigenvectors as COLUMNS.
                # Since H = iA, dU = exp(-A) = exp(iH) = ZDZ^d
                dU = np.dot(Z * np.exp(1.j * epsilon), dag(Z))
                if U.dtype == float:
                    U[:] = np.dot(U, dU).real
                else:
                    U[:] = np.dot(U, dU)

        self.update()



import functools
from math import sqrt
from pathlib import Path
import numpy as np

from ase import Atom, Atoms
from ase.build import bulk, molecule
from ase.lattice.compounds import L1_2
from ase.lattice.hexagonal import Graphene
from gpaw import Davidson, FermiDirac, GPAW, Mixer, PW
from gpaw.mpi import world, serial_comm
from gpaw.new.ase_interface import GPAW as GPAWNew
from gpaw.poisson import FDPoissonSolver, PoissonSolver
from gpaw.test.cachedfilehandler import CachedFilesHandler

response_band_cutoff = {}
_all_gpw_methodnames = set()


def with_band_cutoff(*, gpw, band_cutoff):
    # Store the band cutoffs in a dictionary to aid response tests
    response_band_cutoff[gpw] = band_cutoff

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, band_cutoff=band_cutoff, **kwargs)
        return wrapper

    return decorator


def partial(fun, **kwargs):
    import copy
    kwargs = copy.deepcopy(kwargs)

    def f(self):
        return fun(self, **kwargs)
    return f


def _si_gw(self, *, a, symm, name):
    atoms = self.generate_si_systems()[a]
    return self._si_gw(atoms=atoms,
                       symm=symm,
                       name=f'{name}.txt')


def si_gpwfiles():
    gpw_file_dict = {}
    for a in [0, 1]:
        for symm, name1 in [({}, 'all'), ('off', 'no'),
                            ({'point_group': False}, 'tr'),
                            ({'time_reversal': False}, 'pg')]:
            name = f'si_gw_a{a}_{name1}'
            """
            In !2153, a bug related to late binding of local functions was
            fixed, and the partial wrapper utilized here is a temporary fix.
            Previously, the test would only test the last symmetry in the loop,
            but four times.
            """
            fun = partial(_si_gw, a=a, symm=symm, name=name)
            fun.__name__ = name
            gpw_file_dict[name] = gpwfile(fun)

    return gpw_file_dict


def gpwfile(meth):
    """Decorator to identify the methods that produce gpw files."""
    _all_gpw_methodnames.add(meth.__name__)
    return meth


class GPWFiles(CachedFilesHandler):
    """Create gpw-files."""

    def __init__(self, folder: Path):
        super().__init__(folder, '.gpw')

    def _calculate_and_write(self, name, work_path):
        calc = getattr(self, name)()
        calc.write(work_path, mode='all')

    @gpwfile
    def bcc_li_pw(self):
        return self.bcc_li({'name': 'pw', 'ecut': 200})

    @gpwfile
    def bcc_li_fd(self):
        return self.bcc_li({'name': 'fd'})

    @gpwfile
    def bcc_li_lcao(self):
        return self.bcc_li({'name': 'lcao'})

    def bcc_li(self, mode):
        li = bulk('Li', 'bcc', 3.49)
        li.calc = GPAW(mode=mode,
                       kpts=(3, 3, 3),
                       txt=self.folder / f'bcc_li_{mode["name"]}.txt')
        li.get_potential_energy()
        return li.calc

    @gpwfile
    def be_atom_fd(self):
        atoms = Atoms('Be', [(0, 0, 0)], pbc=False)
        atoms.center(vacuum=6)
        calc = GPAW(mode='fd', h=0.35, symmetry={'point_group': False},
                    txt=self.folder / 'be_atom_fd.txt')
        atoms.calc = calc
        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def fcc_Ni_col(self):
        return self._fcc_Ni('col')

    @gpwfile
    def fcc_Ni_ncol(self):
        return self._fcc_Ni('ncol')

    @gpwfile
    def fcc_Ni_ncolsoc(self):
        return self._fcc_Ni('ncolsoc')

    def _fcc_Ni(self, calc_type):
        Ni = bulk('Ni', 'fcc', 3.48)
        Ni.center()

        mm = 0.5
        easy_axis = 1 / np.sqrt(3) * np.ones(3)
        Ni.set_initial_magnetic_moments([mm])

        symmetry = {'point_group': False, 'time_reversal': True} if \
            calc_type == 'col' else 'off'
        magmoms = None if calc_type == 'col' else [mm * easy_axis]
        soc = True if calc_type == 'ncolsoc' else False

        Ni.calc = GPAWNew(mode={'name': 'pw', 'ecut': 280}, xc='LDA',
                          kpts={'size': (4, 4, 4), 'gamma': True},
                          parallel={'domain': 1, 'band': 1},
                          mixer={'beta': 0.5},
                          symmetry=symmetry,
                          occupations={'name': 'fermi-dirac', 'width': 0.05},
                          convergence={'density': 1e-4},
                          magmoms=magmoms, soc=soc,
                          txt=self.folder / f'fcc_Ni_{calc_type}.txt')
        Ni.get_potential_energy()
        return Ni.calc

    @gpwfile
    def h2_pw(self):
        return self.h2({'name': 'pw', 'ecut': 200})

    @gpwfile
    def h2_fd(self):
        return self.h2({'name': 'fd'})

    @gpwfile
    def h2_lcao(self):
        return self.h2({'name': 'lcao'})

    def h2(self, mode):
        h2 = Atoms('H2', positions=[[0, 0, 0], [0.74, 0, 0]])
        h2.center(vacuum=2.5)
        h2.calc = GPAW(mode=mode,
                       txt=self.folder / f'h2_{mode["name"]}.txt')
        h2.get_potential_energy()
        return h2.calc

    @gpwfile
    def h2_pw_0(self):
        h2 = Atoms('H2',
                   positions=[[-0.37, 0, 0], [0.37, 0, 0]],
                   cell=[5.74, 5, 5],
                   pbc=True)
        h2.calc = GPAW(mode={'name': 'pw', 'ecut': 200},
                       txt=self.folder / 'h2_pw_0.txt')
        h2.get_potential_energy()
        return h2.calc

    @gpwfile
    def h2_bcc_afm(self):
        a = 2.75
        atoms = bulk(name='H', crystalstructure='bcc', a=a, cubic=True)
        atoms.set_initial_magnetic_moments([1., -1.])

        atoms.calc = GPAW(xc='LDA',
                          txt=self.folder / 'h2_bcc_afm.txt',
                          mode=PW(250),
                          nbands=4,
                          convergence={'bands': 4},
                          kpts={'density': 2.0, 'gamma': True})
        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def silicon_pdens_tool(self):
        # used by response code's pdens tool test
        pw = 200
        kpts = 3
        nbands = 8

        a = 5.431
        atoms = bulk('Si', 'diamond', a=a)

        calc = GPAW(mode=PW(pw),
                    kpts=(kpts, kpts, kpts),
                    nbands=nbands,
                    convergence={'bands': -1},
                    xc='LDA',
                    occupations=FermiDirac(0.001))

        atoms.calc = calc
        atoms.get_potential_energy()
        return calc

    @gpwfile
    def h_pw(self):
        h = Atoms('H', magmoms=[1])
        h.center(vacuum=4.0)
        h.calc = GPAW(mode={'name': 'pw', 'ecut': 500},
                      txt=self.folder / 'h_pw.txt')
        h.get_potential_energy()
        return h.calc

    @gpwfile
    def h_chain(self):
        from gpaw.new.ase_interface import GPAW
        a = 2.5
        k = 4
        """Compare 2*H AFM cell with 1*H q=1/2 spin-spiral cell."""
        h = Atoms('H',
                  magmoms=[1],
                  cell=[a, 0, 0],
                  pbc=[1, 0, 0])
        h.center(vacuum=2.0, axis=(1, 2))
        h.calc = GPAW(mode={'name': 'pw',
                            'ecut': 400,
                            'qspiral': [0.5, 0, 0]},
                      magmoms=[[1, 0, 0]],
                      symmetry='off',
                      kpts=(2 * k, 1, 1),
                      txt=self.folder / 'h_chain.txt')
        h.get_potential_energy()
        return h.calc

    @gpwfile
    def h_shift(self):
        # Check for Hydrogen atom
        atoms = Atoms('H', cell=(3 * np.eye(3)), pbc=True)

        # Do a GS and save it
        calc = GPAW(
            mode=PW(600), symmetry={'point_group': False},
            kpts={'size': (2, 2, 2)}, nbands=5, txt=None)
        atoms.calc = calc
        atoms.get_potential_energy()

        return atoms.calc

    @gpwfile
    def h2_chain(self):
        a = 2.5
        k = 4
        h2 = Atoms('H2',
                   [(0, 0, 0), (a, 0, 0)],
                   magmoms=[1, -1],
                   cell=[2 * a, 0, 0],
                   pbc=[1, 0, 0])
        h2.center(vacuum=2.0, axis=(1, 2))
        h2.calc = GPAW(mode={'name': 'pw',
                             'ecut': 400},
                       kpts=(k, 1, 1),
                       txt=self.folder / 'h2_chain.txt')
        h2.get_potential_energy()
        return h2.calc

    @gpwfile
    def h2_lcao_pair(self):
        atoms = molecule('H2')
        atoms.set_cell([6.4, 6.4, 6.4])
        atoms.center()

        atoms.calc = GPAW(mode='lcao', occupations=FermiDirac(0.1),
                          poissonsolver={'name': 'fd'},
                          txt=self.folder / 'h2_lcao_pair.txt')
        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def na_chain(self):
        a = 3
        atom = Atoms('Na',
                     cell=[a, 0, 0],
                     pbc=[1, 1, 1])
        atom.center(vacuum=1 * a, axis=(1, 2))
        atom.center()
        atoms = atom.repeat((2, 1, 1))

        calc = GPAW(mode=PW(200),
                    nbands=4,
                    setups={'Na': '1'},
                    txt=self.folder / 'na_chain.txt',
                    kpts=(16, 2, 2))

        atoms.calc = calc
        atoms.get_potential_energy()
        return calc

    @gpwfile
    def n2_pw(self):
        N2 = molecule('N2')
        N2.center(vacuum=2.0)

        N2.calc = GPAW(mode=PW(force_complex_dtype=True),
                       xc='PBE',
                       parallel={'domain': 1},
                       eigensolver='rmm-diis',
                       txt=self.folder / 'n2_pw.txt')

        N2.get_potential_energy()
        N2.calc.diagonalize_full_hamiltonian(nbands=104, scalapack=True)
        return N2.calc

    @gpwfile
    def n_pw(self):
        N2 = molecule('N2')
        N2.center(vacuum=2.0)

        N = molecule('N')
        N.set_cell(N2.cell)
        N.center()

        N.calc = GPAW(mode=PW(force_complex_dtype=True),
                      xc='PBE',
                      parallel={'domain': 1},
                      eigensolver='rmm-diis',
                      txt=self.folder / 'n_pw.txt')
        N.get_potential_energy()
        N.calc.diagonalize_full_hamiltonian(nbands=104, scalapack=True)
        return N.calc

    @gpwfile
    def o2_pw(self):
        d = 1.1
        a = Atoms('O2', positions=[[0, 0, 0], [d, 0, 0]], magmoms=[1, 1])
        a.center(vacuum=4.0)
        a.calc = GPAW(mode={'name': 'pw', 'ecut': 800},
                      txt=self.folder / 'o2_pw.txt')
        a.get_potential_energy()
        return a.calc

    @gpwfile
    def Cu3Au_qna(self):
        ecut = 300
        kpts = (1, 1, 1)

        QNA = {'alpha': 2.0,
               'name': 'QNA',
               'stencil': 1,
               'orbital_dependent': False,
               'parameters': {'Au': (0.125, 0.1), 'Cu': (0.0795, 0.005)},
               'setup_name': 'PBE',
               'type': 'qna-gga'}

        atoms = L1_2(['Au', 'Cu'], latticeconstant=3.7)
        atoms[0].position[0] += 0.01  # Break symmetry already here
        calc = GPAW(mode=PW(ecut),
                    eigensolver=Davidson(2),
                    nbands='120%',
                    mixer=Mixer(0.4, 7, 50.0),
                    parallel=dict(domain=1),
                    convergence={'density': 1e-4},
                    xc=QNA,
                    kpts=kpts,
                    txt=self.folder / 'Cu3Au_qna.txt')
        atoms.calc = calc
        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def co_mom(self):
        atoms = molecule('CO')
        atoms.center(vacuum=2)

        calc = GPAW(mode='lcao',
                    basis='dzp',
                    nbands=7,
                    h=0.24,
                    xc='PBE',
                    spinpol=True,
                    symmetry='off',
                    convergence={'energy': 100,
                                 'density': 1e-3},
                    txt=self.folder / 'co_mom.txt')

        atoms.calc = calc
        # Ground-state calculation
        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def co_lcao(self):
        d = 1.1
        co = Atoms('CO', positions=[[0, 0, 0], [d, 0, 0]])
        co.center(vacuum=4.0)
        co.calc = GPAW(mode='lcao',
                       txt=self.folder / 'co_lcao.txt')
        co.get_potential_energy()
        return co.calc

    def _c2h4(self):
        d = 1.54
        h = 1.1
        x = d * (2 / 3)**0.5
        z = d / 3**0.5
        pe = Atoms('C2H4',
                   positions=[[0, 0, 0],
                              [x, 0, z],
                              [0, -h * (2 / 3)**0.5, -h / 3**0.5],
                              [0, h * (2 / 3)**0.5, -h / 3**0.5],
                              [x, -h * (2 / 3)**0.5, z + h / 3**0.5],
                              [x, h * (2 / 3)**0.5, z + h / 3**0.5]],
                   cell=[2 * x, 0, 0],
                   pbc=(1, 0, 0))
        pe.center(vacuum=2.0, axis=(1, 2))
        return pe

    def c2h4_pw_nosym(self):
        pe = self._c2h4()
        pe.calc = GPAW(mode='pw',
                       kpts=(3, 1, 1),
                       symmetry='off',
                       txt=self.folder / 'c2h4_pw_nosym.txt')
        pe.get_potential_energy()
        return pe.calc

    @gpwfile
    def c6h12_pw(self):
        pe = self._c2h4()
        pe = pe.repeat((3, 1, 1))
        pe.calc = GPAW(mode='pw', txt=self.folder / 'c6h12_pw.txt')
        pe.get_potential_energy()
        return pe.calc

    @gpwfile
    def h2o_lcao(self):
        atoms = molecule('H2O', cell=[8, 8, 8], pbc=1)
        atoms.center()
        atoms.calc = GPAW(mode='lcao', txt=self.folder / 'h2o_lcao.txt')
        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def h2o_xas(self):
        from math import cos, pi, sin

        setupname = 'h2o_xas_hch1s'
        self.generator2_setup(
            'O', 8, '2s,s,2p,p,d', [1.2], 1.0, None, 2,
            core_hole='1s,0.5',
            name=setupname)

        a = 5.0
        d = 0.9575
        t = pi / 180 * 104.51
        H2O = Atoms(
            [
                Atom('O', (0, 0, 0)),
                Atom('H', (d, 0, 0)),
                Atom('H', (d * cos(t), d * sin(t), 0)),
            ],
            cell=(a, a, a),
            pbc=False,
        )
        H2O.center()
        calc = GPAW(
            txt=self.folder / 'h2o_xas.txt',
            mode='fd',
            nbands=10,
            h=0.2,
            setups={'O': 'h2o_xas_hch1s'},
            poissonsolver=FDPoissonSolver(use_charge_center=True),
        )
        H2O.calc = calc
        _ = H2O.get_potential_energy()
        return calc

    @gpwfile
    def h20_lr2_nbands8(self):
        return self.h2o_nbands(
            {'poissonsolver': {'name': 'fd'}, 'nbands': 8, 'basis': 'dzp'})

    @gpwfile
    def h20_lr2_nbands6(self):
        return self.h2o_nbands({'nbands': 6, 'basis': 'sz(dzp)'})

    def h2o_nbands(self, params):
        atoms = molecule('H2O')
        atoms.center(vacuum=4)

        calc = GPAW(h=0.4, **params, xc='LDA', mode='lcao',
                    txt=self.folder / f'h2o_lr2_nbands{params["nbands"]}.out')
        atoms.calc = calc
        atoms.get_potential_energy()

        return atoms.calc

    @gpwfile
    def si_fd_ibz(self):
        si = bulk('Si', 'diamond', a=5.43)
        k = 3
        si.calc = GPAW(mode='fd', kpts=(k, k, k),
                       txt=self.folder / 'si_fd_ibz.txt')
        si.get_potential_energy()
        return si.calc

    @gpwfile
    def si_fd_bz(self):
        si = bulk('Si', 'diamond', a=5.43)
        k = 3
        si.calc = GPAW(mode='fd', kpts=(k, k, k,),
                       symmetry={'point_group': False,
                                 'time_reversal': False},
                       txt=self.folder / 'si_fd_bz.txt')
        si.get_potential_energy()
        return si.calc

    @gpwfile
    def si_pw(self):
        si = bulk('Si')
        calc = GPAW(mode='pw',
                    xc='LDA',
                    occupations=FermiDirac(width=0.001),
                    kpts={'size': (2, 2, 2), 'gamma': True},
                    txt=self.folder / 'si_pw.txt')
        si.calc = calc
        si.get_potential_energy()
        return si.calc

    @gpwfile
    def si_qpoint_rounding_bug(self):
        # Test system for guarding against inconsistent kpoints as in #1178.
        from ase.calculators.calculator import kpts2kpts
        atoms = bulk('Si')
        kpts = kpts2kpts(kpts={'gamma': True, 'size': (6, 1, 1)}, atoms=atoms)

        # Error happened when qpoint was ~1e-17 yet was not considered gamma.
        # Add a bit of noise on purpose so we are sure to hit such a case,
        # even if the underlying implementation changes:
        kpts.kpts += np.linspace(1e-16, 1e-15, 18).reshape(6, 3)

        calc = GPAW(mode=PW(340), kpts=kpts)
        atoms.calc = calc
        atoms.get_potential_energy()
        return calc

    @property
    def testing_setup_path(self):
        # Some calculations in gpwfile fixture like to use funny setups.
        # This is not so robust since the setups will be all jumbled.
        # We could improve the mechanism by programmatic naming/subfolders.
        return self.folder / 'setups'

    def save_setup(self, setup):
        self.testing_setup_path.mkdir(parents=True, exist_ok=True)
        setup_file = self.testing_setup_path / setup.stdfilename
        if world.rank == 0:
            setup.write_xml(setup_file)
        world.barrier()
        return setup

    def generate_setup(self, *args, **kwargs):
        from gpaw.test import gen
        setup = gen(*args, **kwargs, write_xml=False)
        self.save_setup(setup)
        return setup

    def generator2_setup(self, *args, name, **kwargs):
        from gpaw.atom.generator2 import generate
        gen = generate(*args, **kwargs)
        setup = gen.make_paw_setup(name)
        self.save_setup(setup)
        return setup

    @gpwfile
    def si_corehole_pw(self):
        # Generate setup for oxygen with half a core-hole:
        setupname = 'si_corehole_pw_hch1s'
        self.generate_setup('Si', name=setupname,
                            corehole=(1, 0, 0.5), gpernode=30)

        a = 2.6
        si = Atoms('Si', cell=(a, a, a), pbc=True)

        calc = GPAW(mode='fd',
                    txt=self.folder / 'si_corehole_pw.txt',
                    nbands=None,
                    h=0.25,
                    occupations=FermiDirac(width=0.05),
                    setups='si_corehole_pw_hch1s',
                    convergence={'maximum iterations': 1})
        si.calc = calc
        _ = si.get_potential_energy()
        return si.calc

    @gpwfile
    def si_corehole_sym_pw(self):
        setupname = 'si_corehole_sym_pw_hch1s'
        self.generate_setup('Si', name=setupname, corehole=(1, 0, 0.5),
                            gpernode=30)
        return self.si_corehole_sym(sym={}, setupname=setupname)

    @gpwfile
    def si_corehole_nosym_pw(self):
        setupname = 'si_corehole_sym_pw_hch1s'
        # XXX same setup as above, but we have it twice since caching
        # works per gpw file and not per setup
        self.generate_setup('Si', name=setupname, corehole=(1, 0, 0.5),
                            gpernode=30)
        return self.si_corehole_sym(sym='off', setupname=setupname)

    def si_corehole_sym(self, sym, setupname):
        tag = 'nosym' if sym == 'off' else 'sym'

        a = 5.43095
        si_nonortho = Atoms(
            [Atom('Si', (0, 0, 0)), Atom('Si', (a / 4, a / 4, a / 4))],
            cell=[(a / 2, a / 2, 0), (a / 2, 0, a / 2), (0, a / 2, a / 2)],
            pbc=True,
        )
        # calculation with full symmetry
        calc = GPAW(
            mode='fd',
            txt=self.folder / f'si_corehole_{tag}_hch1s.txt',
            nbands=-10,
            h=0.25,
            kpts=(2, 2, 2),
            occupations=FermiDirac(width=0.05),
            setups={0: setupname},
            symmetry=sym
        )
        si_nonortho.calc = calc
        _ = si_nonortho.get_potential_energy()
        return calc

    @gpwfile
    @with_band_cutoff(gpw='fancy_si_pw',
                      band_cutoff=8)  # 2 * (3s, 3p)
    def _fancy_si(self, *, band_cutoff, symmetry=None):
        if symmetry is None:
            symmetry = {}
        xc = 'LDA'
        kpts = 4
        pw = 300
        occw = 0.01
        conv = {'bands': band_cutoff + 1,
                'density': 1.e-8}
        atoms = bulk('Si')
        atoms.center()

        tag = '_nosym' if symmetry == 'off' else ''
        atoms.calc = GPAW(
            xc=xc,
            mode=PW(pw),
            kpts={'size': (kpts, kpts, kpts), 'gamma': True},
            nbands=band_cutoff + 12,  # + 2 * (3s, 3p),
            occupations=FermiDirac(occw),
            convergence=conv,
            txt=self.folder / f'fancy_si_pw{tag}.txt',
            symmetry=symmetry)

        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def fancy_si_pw(self):
        return self._fancy_si()

    @gpwfile
    def fancy_si_pw_nosym(self):
        return self._fancy_si(symmetry='off')

    @with_band_cutoff(gpw='sic_pw',
                      band_cutoff=8)  # (3s, 3p) + (2s, 2p)
    def _sic_pw(self, *, band_cutoff, spinpol=False):
        """Simple semi-conductor with broken inversion symmetry."""
        # Use the diamond crystal structure as blue print
        diamond = bulk('C', 'diamond')
        si = bulk('Si', 'diamond')
        # Break inversion symmetry by substituting one Si for C
        atoms = si.copy()
        atoms.symbols = 'CSi'
        # Scale the cell to the diamond/Si average
        cell_cv = (diamond.get_cell() + si.get_cell()) / 2.
        atoms.set_cell(cell_cv)

        # Set up calculator
        tag = '_spinpol' if spinpol else ''
        atoms.calc = GPAW(
            mode=PW(400),
            xc='LDA',
            kpts={'size': (4, 4, 4)},
            symmetry={'point_group': False,
                      'time_reversal': True},
            nbands=band_cutoff + 6,
            occupations=FermiDirac(0.001),
            convergence={'bands': band_cutoff + 1,
                         'density': 1e-8},
            spinpol=spinpol,
            txt=self.folder / f'sic_pw{tag}.txt'
        )

        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def sic_pw(self):
        return self._sic_pw()

    @gpwfile
    def sic_pw_spinpol(self):
        return self._sic_pw(spinpol=True)

    @staticmethod
    def generate_si_systems():
        a = 5.43
        si1 = bulk('Si', 'diamond', a=a)
        si2 = si1.copy()
        si2.positions -= a / 8
        return [si1, si2]

    def _si_gw(self, atoms, symm, name):
        atoms.calc = GPAW(mode=PW(250),
                          eigensolver='rmm-diis',
                          occupations=FermiDirac(0.01),
                          symmetry=symm,
                          kpts={'size': (2, 2, 2), 'gamma': True},
                          convergence={'density': 1e-7},
                          parallel={'domain': 1},
                          txt=self.folder / name)
        atoms.get_potential_energy()
        scalapack = atoms.calc.wfs.bd.comm.size
        atoms.calc.diagonalize_full_hamiltonian(nbands=8, scalapack=scalapack)
        return atoms.calc

    @gpwfile
    def na_pw(self):
        from ase.build import bulk

        blk = bulk('Na', 'bcc', a=4.23)

        ecut = 350
        blk.calc = GPAW(mode=PW(ecut),
                        basis='dzp',
                        kpts={'size': (4, 4, 4), 'gamma': True},
                        parallel={'domain': 1, 'band': 1},
                        txt=self.folder / 'na_pw.txt',
                        nbands=4,
                        occupations=FermiDirac(0.01),
                        setups={'Na': '1'})
        blk.get_potential_energy()
        blk.calc.diagonalize_full_hamiltonian(nbands=520)
        return blk.calc

    @gpwfile
    def na2_tddft_poisson_sym(self):
        return self._na2_tddft(name='poisson_sym', basis='dzp', xc='LDA')

    @gpwfile
    def na2_tddft_poisson(self):
        return self._na2_tddft(name='poisson', basis='dzp', xc='LDA')

    @gpwfile
    def na2_tddft_dzp(self):
        return self._na2_tddft(name='dzp', basis='dzp', xc='LDA')

    @gpwfile
    def na2_tddft_sz(self):
        return self._na2_tddft(name='sz', basis='sz(dzp)', xc='oldLDA')

    def _na2_tddft(self, name, basis, xc):
        atoms = molecule('Na2')
        atoms.center(vacuum=4.0)

        names = ['poisson_sym', 'poisson']
        poisson = PoissonSolver('fd', eps=1e-16) if name in names \
            else None
        symmetry = {'point_group': True if name == 'poisson_sym' else False}

        calc = GPAW(nbands=2, h=0.4, setups=dict(Na='1'),
                    basis=basis, mode='lcao', xc=xc,
                    convergence={'density': 1e-8},
                    poissonsolver=poisson,
                    communicator=serial_comm if xc == 'oldLDA' else world,
                    symmetry=symmetry,
                    txt=self.folder / f'na2_tddft_{name}.out')
        atoms.calc = calc
        atoms.get_potential_energy()

        return calc

    @gpwfile
    def na2_fd(self):
        """Sodium dimer, Na2."""
        d = 1.5
        atoms = Atoms(symbols='Na2',
                      positions=[(0, 0, d),
                                 (0, 0, -d)],
                      pbc=False)

        atoms.center(vacuum=6.0)
        # Larger grid spacing, LDA is ok
        gs_calc = GPAW(mode='fd',
                       txt=self.folder / 'na2_fd.txt',
                       nbands=1, h=0.35, xc='LDA',
                       setups={'Na': '1'},
                       symmetry={'point_group': False})
        atoms.calc = gs_calc
        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def na2_fd_with_sym(self):
        """Sodium dimer, Na2."""
        d = 1.5
        atoms = Atoms(symbols='Na2',
                      positions=[(0, 0, d),
                                 (0, 0, -d)],
                      pbc=False)

        atoms.center(vacuum=6.0)
        # Larger grid spacing, LDA is ok
        gs_calc = GPAW(mode='fd', nbands=1, h=0.35, xc='LDA',
                       txt=self.folder / 'na2_fd_with_sym.txt',
                       setups={'Na': '1'})
        atoms.calc = gs_calc
        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def na2_isolated(self):
        # Permittivity file
        if world.rank == 0:
            fo = open('ed.txt', 'w')
            fo.writelines(['1.20 0.20 25.0'])
            fo.close()
        world.barrier()

        from gpaw.fdtd.poisson_fdtd import FDTDPoissonSolver
        from gpaw.fdtd.polarizable_material import (
            PermittivityPlus,
            PolarizableMaterial,
            PolarizableSphere,
        )

        # Whole simulation cell (Angstroms)
        large_cell = [20, 20, 30]

        # Quantum subsystem
        atom_center = np.array([10.0, 10.0, 20.0])
        atoms = Atoms('Na2', [atom_center + [0.0, 0.0, -1.50],
                              atom_center + [0.0, 0.0, +1.50]])

        # Classical subsystem
        classical_material = PolarizableMaterial()
        sphere_center = np.array([10.0, 10.0, 10.0])
        classical_material.add_component(
            PolarizableSphere(
                permittivity=PermittivityPlus('ed.txt'),
                center=sphere_center,
                radius=5.0
            )
        )

        # Accuracy
        energy_eps = 0.0005
        density_eps = 1e-6
        poisson_eps = 1e-12

        # Combined Poisson solver
        poissonsolver = FDTDPoissonSolver(
            classical_material=classical_material,
            eps=poisson_eps,
            qm_spacing=0.40,
            cl_spacing=0.40 * 4,
            cell=large_cell,
            remove_moments=(1, 4),
            communicator=world,
            potential_coupler='Refiner',
        )
        poissonsolver.set_calculation_mode('iterate')

        # Combined system
        atoms.set_cell(large_cell)
        atoms, qm_spacing, gpts = poissonsolver.cut_cell(atoms, vacuum=2.50)

        # Initialize GPAW
        gs_calc = GPAW(mode='fd',
                       txt=self.folder / 'na2_isolated.txt',
                       gpts=gpts,
                       eigensolver='cg',
                       nbands=-1,
                       poissonsolver=poissonsolver,
                       symmetry={'point_group': False},
                       convergence={'energy': energy_eps,
                                    'density': density_eps})
        atoms.calc = gs_calc

        # Ground state
        atoms.get_potential_energy()

        return gs_calc

    @gpwfile
    def na3_pw_restart(self):
        params = dict(mode=PW(200), convergence={
            'eigenstates': 1.24, 'energy': 2e-1, 'density': 1e-1})
        return self._na3_restart(params=params)

    @gpwfile
    def na3_fd_restart(self):
        params = dict(mode='fd', h=0.30, convergence={
            'eigenstates': 1.24, 'energy': 2e-1, 'density': 1e-1})
        return self._na3_restart(params=params)

    @gpwfile
    def na3_fd_kp_restart(self):
        params = dict(mode='fd', h=0.30, kpts=(1, 1, 3), convergence={
            'eigenstates': 1.24, 'energy': 2e-1, 'density': 1e-1})
        return self._na3_restart(params=params)

    @gpwfile
    def na3_fd_density_restart(self):
        params = dict(mode='fd', h=0.30, convergence={
            'eigenstates': 1.e-3, 'energy': 2e-1, 'density': 1e-1})
        return self._na3_restart(params=params)

    def _na3_restart(self, params):
        d = 3.0
        atoms = Atoms('Na3',
                      positions=[(0, 0, 0),
                                 (0, 0, d),
                                 (0, d * sqrt(3 / 4), d / 2)],
                      magmoms=[1.0, 1.0, 1.0],
                      cell=(3.5, 3.5, 4 + 2 / 3),
                      pbc=True)

        atoms.calc = GPAW(nbands=3,
                          setups={'Na': '1'},
                          **params)
        atoms.get_potential_energy()

        return atoms.calc

    @gpwfile
    def sih4_xc_gllbsc_lcao(self):
        return self._sih4_gllbsc(mode='lcao', basis='dzp')

    @gpwfile
    def sih4_xc_gllbsc_fd(self):
        return self._sih4_gllbsc(mode='fd', basis={})

    def _sih4_gllbsc(self, mode, basis):
        atoms = molecule('SiH4')
        atoms.center(vacuum=4.0)

        # Ground-state calculation
        calc = GPAW(mode=mode, nbands=7, h=0.4,
                    convergence={'density': 1e-8},
                    xc='GLLBSC',
                    basis=basis,
                    symmetry={'point_group': False},
                    txt=self.folder / f'sih4_xc_gllbsc_{mode}.txt')
        atoms.calc = calc
        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def nacl_nospin(self):
        return self._nacl_mol(spinpol=False)

    @gpwfile
    def nacl_spin(self):
        return self._nacl_mol(spinpol=True)

    def _nacl_mol(self, spinpol):
        atoms = molecule('NaCl')
        atoms.center(vacuum=4.0)
        calc = GPAW(nbands=6,
                    h=0.4,
                    setups=dict(Na='1'),
                    basis='dzp',
                    mode='lcao',
                    convergence={'density': 1e-8},
                    spinpol=spinpol,
                    communicator=world,
                    symmetry={'point_group': False},
                    txt='gs.out')
        atoms.calc = calc
        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def nacl_fd(self):
        d = 4.0
        atoms = Atoms('NaCl', [(0, 0, 0), (0, 0, d)])
        atoms.center(vacuum=4.5)

        gs_calc = GPAW(
            txt=self.folder / 'nacl_fd.txt',
            mode='fd', nbands=4, eigensolver='cg',
            gpts=(32, 32, 44), xc='LDA', symmetry={'point_group': False},
            setups={'Na': '1'})
        atoms.calc = gs_calc
        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def bn_pw(self):
        atoms = bulk('BN', 'zincblende', a=3.615)
        atoms.calc = GPAW(mode=PW(400),
                          kpts={'size': (2, 2, 2), 'gamma': True},
                          nbands=12,
                          convergence={'bands': 9},
                          occupations=FermiDirac(0.001),
                          txt=self.folder / 'bn_pw.txt')
        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def hbn_pw(self):
        atoms = Graphene(symbol='B',
                         latticeconstant={'a': 2.5, 'c': 1.0},
                         size=(1, 1, 1))
        atoms[0].symbol = 'N'
        atoms.pbc = (1, 1, 0)
        atoms.center(axis=2, vacuum=3.0)
        atoms.calc = GPAW(txt=self.folder / 'hbn_pw.txt',
                          mode=PW(400),
                          xc='LDA',
                          nbands=50,
                          occupations=FermiDirac(0.001),
                          parallel={'domain': 1},
                          convergence={'bands': 26},
                          kpts={'size': (3, 3, 1), 'gamma': True})
        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def graphene_pw(self):
        from ase.lattice.hexagonal import Graphene
        atoms = Graphene(symbol='C',
                         latticeconstant={'a': 2.45, 'c': 1.0},
                         size=(1, 1, 1))
        atoms.pbc = (1, 1, 0)
        atoms.center(axis=2, vacuum=4.0)
        ecut = 250
        nkpts = 6
        atoms.calc = GPAW(mode=PW(ecut),
                          kpts={'size': (nkpts, nkpts, 1), 'gamma': True},
                          nbands=len(atoms) * 6,
                          txt=self.folder / 'graphene_pw.txt')
        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def i2sb2_pw_nosym(self):
        # Structure from c2db
        atoms = Atoms('I2Sb2',
                      positions=[[0.02437357, 0.05048655, 6.11612164],
                                 [0.02524896, 3.07135573, 11.64646853],
                                 [0.02717742, 0.01556495, 8.89278807],
                                 [0.02841809, 3.10675382, 8.86983839]],
                      cell=[[5.055642258802973, -9.89475498615942e-15, 0.0],
                            [-2.5278211265136266, 4.731999711338355, 0.0],
                            [3.38028806436979e-15, 0.0, 18.85580293064]],
                      pbc=(1, 1, 0))
        atoms.calc = GPAW(mode=PW(250),
                          xc='PBE',
                          kpts={'size': (6, 6, 1), 'gamma': True},
                          txt=self.folder / 'i2sb2_pw_nosym.txt',
                          symmetry='off')

        atoms.get_potential_energy()
        return atoms.calc

    @with_band_cutoff(gpw='bi2i6_pw',
                      band_cutoff=36)
    def _bi2i6(self, *, band_cutoff, symmetry=None):
        if symmetry is None:
            symmetry = {}
        positions = [[4.13843656, 2.38932746, 9.36037077],
                     [0.00000000, 4.77865492, 9.36034750],
                     [3.89827619, 0.00000000, 7.33713295],
                     [2.18929748, 3.79197674, 7.33713295],
                     [-1.94913711, 3.37600678, 7.33713295],
                     [3.89827619, 0.00000000, 11.3835853],
                     [2.18929961, 3.79197551, 11.3835853],
                     [-1.94913924, 3.37600555, 11.3835853]]
        cell = [[8.276873113486648, 0.0, 0.0],
                [-4.138436556743325, 7.167982380179831, 0.0],
                [0.0, 0.0, 0.0]]
        pbc = [True, True, False]
        atoms = Atoms('Bi2I6',
                      positions=positions,
                      cell=cell,
                      pbc=pbc)
        atoms.center(vacuum=3.0, axis=2)

        ecut = 120
        nkpts = 4
        conv = {'bands': band_cutoff + 1,
                'density': 1.e-4}

        tag = '_nosym' if symmetry == 'off' else ''
        atoms.calc = GPAW(mode=PW(ecut),
                          xc='LDA',
                          kpts={'size': (nkpts, nkpts, 1), 'gamma': True},
                          occupations=FermiDirac(0.01),
                          mixer={'beta': 0.5},
                          convergence=conv,
                          nbands=band_cutoff + 9,
                          txt=self.folder / f'bi2i6_pw{tag}.txt',
                          symmetry=symmetry)

        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def bi2i6_pw(self):
        return self._bi2i6()

    @gpwfile
    def bi2i6_pw_nosym(self):
        return self._bi2i6(symmetry='off')

    def _mos2(self, symmetry=None):
        if symmetry is None:
            symmetry = {}
        from ase.build import mx2
        atoms = mx2(formula='MoS2', kind='2H', a=3.184, thickness=3.127,
                    size=(1, 1, 1), vacuum=5)
        atoms.pbc = (1, 1, 0)
        ecut = 250
        nkpts = 6
        tag = '_nosym' if symmetry == 'off' else ''
        atoms.calc = GPAW(mode=PW(ecut),
                          xc='LDA',
                          kpts={'size': (nkpts, nkpts, 1), 'gamma': True},
                          occupations=FermiDirac(0.01),
                          txt=self.folder / f'mos2_pw{tag}.txt',
                          symmetry=symmetry)

        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def mos2_pw(self):
        return self._mos2()

    @gpwfile
    def mos2_pw_nosym(self):
        return self._mos2(symmetry='off')

    @gpwfile
    def mos2_5x5_pw(self):
        calc = GPAW(mode=PW(180),
                    xc='PBE',
                    nbands='nao',
                    setups={'Mo': '6'},
                    occupations=FermiDirac(0.001),
                    convergence={'bands': -5},
                    kpts=(5, 5, 1))

        from ase.build import mx2
        layer = mx2(formula='MoS2', kind='2H', a=3.1604, thickness=3.172,
                    size=(1, 1, 1), vacuum=3.414)
        layer.pbc = (1, 1, 0)
        layer.calc = calc
        layer.get_potential_energy()
        return layer.calc

    @with_band_cutoff(gpw='p4_pw',
                      band_cutoff=40)
    def _p4(self, band_cutoff, spinpol=False):
        atoms = Atoms('P4', positions=[[0.03948480, -0.00027057, 7.49990646],
                                       [0.86217564, -0.00026338, 9.60988536],
                                       [2.35547782, 1.65277230, 9.60988532],
                                       [3.17816857, 1.65277948, 7.49990643]],
                      cell=[4.63138807675, 3.306178252090, 17.10979291],
                      pbc=[True, True, False])
        atoms.center(vacuum=1.5, axis=2)
        tag = '_spinpol' if spinpol else ''
        nkpts = 2
        atoms.calc = GPAW(mode=PW(250),
                          xc='LDA', spinpol=spinpol,
                          kpts={'size': (nkpts, nkpts, 1), 'gamma': True},
                          occupations={'width': 0},
                          nbands=band_cutoff + 10,
                          convergence={'bands': band_cutoff + 1},
                          txt=self.folder / f'p4_pw{tag}.txt')
        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def p4_pw(self):
        return self._p4()

    @gpwfile
    def p4_pw_spinpol(self):
        return self._p4(spinpol=True)

    @gpwfile
    def ni_pw_kpts333(self):
        from ase.dft.kpoints import monkhorst_pack
        # from gpaw.mpi import serial_comm
        Ni = bulk('Ni', 'fcc')
        Ni.set_initial_magnetic_moments([0.7])

        kpts = monkhorst_pack((3, 3, 3))

        calc = GPAW(mode='pw',
                    txt=self.folder / 'ni_pw_kpts333.txt',
                    kpts=kpts,
                    occupations=FermiDirac(0.001),
                    setups={'Ni': '10'},
                    parallel=dict(domain=1),  # >1 fails on 8 cores
                    # communicator=serial_comm
                    )

        Ni.calc = calc
        Ni.get_potential_energy()
        calc.diagonalize_full_hamiltonian()
        return calc

    @gpwfile
    def c_pw(self):
        atoms = bulk('C')
        atoms.center()
        calc = GPAW(mode=PW(150),
                    txt=self.folder / 'c_pw.txt',
                    convergence={'bands': 6},
                    nbands=12,
                    kpts={'gamma': True, 'size': (2, 2, 2)},
                    xc='LDA')

        atoms.calc = calc
        atoms.get_potential_energy()
        return atoms.calc

    def _nicl2_pw(self, vacuum=3.0, identifier='', ecut=285, **kwargs):
        from ase.build import mx2

        # Define input parameters
        kpts = 6
        occw = 0.01
        conv = {'density': 1.e-3}

        a = 3.502
        thickness = 2.617
        mm = 2.0

        # Set up atoms
        atoms = mx2(formula='NiCl2', kind='1T', a=a,
                    thickness=thickness, vacuum=vacuum)
        atoms.set_initial_magnetic_moments([mm, 0.0, 0.0])
        # Use pbc to allow for real-space density interpolation
        atoms.pbc = True

        dct = dict(
            mixer={'beta': 0.75, 'nmaxold': 8, 'weight': 100.0},
            mode=PW(ecut,
                    # Interpolate the density in real-space
                    interpolation=3),
            kpts={'size': (kpts, kpts, 1), 'gamma': True},
            occupations=FermiDirac(occw),
            convergence=conv,
            txt=self.folder / f'nicl2_pw{identifier}.txt')
        dct.update(kwargs)
        atoms.calc = GPAW(**dct)

        atoms.get_potential_energy()

        return atoms.calc

    @gpwfile
    def nicl2_pw(self):
        return self._nicl2_pw(vacuum=3.0, ecut=300.0)

    @gpwfile
    def nicl2_pw_evac(self):
        return self._nicl2_pw(vacuum=10.0, identifier='_evac')

    @gpwfile
    def Tl_box_pw(self):
        Tl = Atoms('Tl',
                   cell=[5, 5, 5],
                   scaled_positions=[[0.5, 0.5, 0.5]],
                   pbc=False)

        Tl.calc = GPAWNew(mode={'name': 'pw', 'ecut': 300}, xc='LDA',
                          occupations={'name': 'fermi-dirac', 'width': 0.01},
                          symmetry='off',
                          convergence={'density': 1e-6},
                          magmoms=[[0, 0, 0.5]], soc=True,
                          txt=self.folder / 'Tl_box_pw.txt')
        Tl.get_potential_energy()
        return Tl.calc

    @with_band_cutoff(gpw='v2br4_pw',
                      band_cutoff=28)  # V(4s,3d) = 6, Br(4s,4p) = 4
    def _v2br4(self, *, band_cutoff, symmetry=None):
        from ase.build import mx2

        if symmetry is None:
            symmetry = {}

        # Define input parameters
        xc = 'LDA'
        kpts = 4
        pw = 200
        occw = 0.01
        conv = {'density': 1.e-4,
                'bands': band_cutoff + 1}

        a = 3.840
        thickness = 2.897
        vacuum = 3.0
        mm = 3.0

        # Set up atoms
        atoms = mx2(formula='VBr2', kind='1T', a=a,
                    thickness=thickness, vacuum=vacuum)
        atoms = atoms.repeat((1, 2, 1))
        atoms.set_initial_magnetic_moments([mm, 0.0, 0.0, -mm, 0.0, 0.0])
        # Use pbc to allow for real-space density interpolation
        atoms.pbc = True

        # Set up calculator
        tag = '_nosym' if symmetry == 'off' else ''
        atoms.calc = GPAW(
            xc=xc,
            mode=PW(pw,
                    # Interpolate the density in real-space
                    interpolation=3),
            kpts={'size': (kpts, kpts // 2, 1), 'gamma': True},
            mixer={'beta': 0.5},
            setups={'V': '5'},
            nbands=band_cutoff + 12,
            occupations=FermiDirac(occw),
            convergence=conv,
            symmetry=symmetry,
            txt=self.folder / f'v2br4_pw{tag}.txt')

        atoms.get_potential_energy()

        return atoms.calc

    @gpwfile
    def v2br4_pw(self):
        return self._v2br4()

    @gpwfile
    def v2br4_pw_nosym(self):
        return self._v2br4(symmetry='off')

    @with_band_cutoff(gpw='fe_pw',
                      band_cutoff=9)  # 4s, 4p, 3d = 9
    def _fe(self, *, band_cutoff, symmetry=None):
        if symmetry is None:
            symmetry = {}
        """See also the fe_fixture_test.py test."""
        xc = 'LDA'
        kpts = 4
        pw = 300
        occw = 0.01
        conv = {'bands': band_cutoff + 1,
                'density': 1.e-8}
        a = 2.867
        mm = 2.21
        atoms = bulk('Fe', 'bcc', a=a)
        atoms.set_initial_magnetic_moments([mm])
        atoms.center()
        tag = '_nosym' if symmetry == 'off' else ''

        atoms.calc = GPAW(
            xc=xc,
            mode=PW(pw),
            kpts={'size': (kpts, kpts, kpts)},
            nbands=band_cutoff + 9,
            occupations=FermiDirac(occw),
            convergence=conv,
            txt=self.folder / f'fe_pw{tag}.txt',
            symmetry=symmetry)

        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def fe_pw(self):
        return self._fe()

    @gpwfile
    def fe_pw_nosym(self):
        return self._fe(symmetry='off')

    @with_band_cutoff(gpw='co_pw',
                      band_cutoff=14)  # 2 * (4s + 3d)
    def _co(self, *, band_cutoff, symmetry=None):
        if symmetry is None:
            symmetry = {}
        # ---------- Inputs ---------- #

        # Atomic configuration
        a = 2.5071
        c = 4.0695
        mm = 1.6
        atoms = bulk('Co', 'hcp', a=a, c=c)
        atoms.set_initial_magnetic_moments([mm, mm])
        atoms.center()

        # Ground state parameters
        xc = 'LDA'
        occw = 0.01
        ebands = 2 * 2  # extra bands for ground state calculation
        pw = 200
        conv = {'density': 1e-8,
                'forces': 1e-8,
                'bands': band_cutoff + 1}

        # ---------- Calculation ---------- #

        tag = '_nosym' if symmetry == 'off' else ''
        atoms.calc = GPAW(xc=xc,
                          mode=PW(pw),
                          kpts={'size': (4, 4, 4), 'gamma': True},
                          occupations=FermiDirac(occw),
                          convergence=conv,
                          nbands=band_cutoff + ebands,
                          symmetry=symmetry,
                          txt=self.folder / f'co_pw{tag}.txt')

        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def co_pw(self):
        return self._co()

    @gpwfile
    def co_pw_nosym(self):
        return self._co(symmetry='off')

    @with_band_cutoff(gpw='srvo3_pw',
                      band_cutoff=20)
    def _srvo3(self, *, band_cutoff, symmetry=None):
        if symmetry is None:
            symmetry = {}

        nk = 3
        cell = bulk('V', 'sc', a=3.901).cell
        atoms = Atoms('SrVO3', cell=cell, pbc=True,
                      scaled_positions=((0.5, 0.5, 0.5),
                                        (0, 0, 0),
                                        (0, 0.5, 0),
                                        (0, 0, 0.5),
                                        (0.5, 0, 0)))
        # Ground state parameters
        xc = 'LDA'
        occw = 0.01
        ebands = 10  # extra bands for ground state calculation
        pw = 200
        conv = {'density': 1e-8,
                'bands': band_cutoff + 1}

        # ---------- Calculation ---------- #

        tag = '_nosym' if symmetry == 'off' else ''
        atoms.calc = GPAW(xc=xc,
                          mode=PW(pw),
                          kpts={'size': (nk, nk, nk), 'gamma': True},
                          occupations=FermiDirac(occw),
                          convergence=conv,
                          nbands=band_cutoff + ebands,
                          symmetry=symmetry,
                          txt=self.folder / f'srvo3_pw{tag}.txt')

        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def srvo3_pw(self):
        return self._srvo3()

    @gpwfile
    def srvo3_pw_nosym(self):
        return self._srvo3(symmetry='off')

    @with_band_cutoff(gpw='al_pw',
                      band_cutoff=10)  # 3s, 3p, 4s, 3d
    def _al(self, *, band_cutoff, symmetry=None):
        if symmetry is None:
            symmetry = {}
        xc = 'LDA'
        kpts = 4
        pw = 300
        occw = 0.01
        conv = {'bands': band_cutoff + 1,
                'density': 1.e-8}
        a = 4.043
        atoms = bulk('Al', 'fcc', a=a)
        atoms.center()
        tag = '_nosym' if symmetry == 'off' else ''

        atoms.calc = GPAW(
            xc=xc,
            mode=PW(pw),
            kpts={'size': (kpts, kpts, kpts), 'gamma': True},
            nbands=band_cutoff + 4,  # + 4p, 5s
            occupations=FermiDirac(occw),
            convergence=conv,
            txt=self.folder / f'al_pw{tag}.txt',
            symmetry=symmetry)

        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def al_pw(self):
        return self._al()

    @gpwfile
    def al_pw_nosym(self):
        return self._al(symmetry='off')

    @gpwfile
    def bse_al(self):
        a = 4.043
        atoms = bulk('Al', 'fcc', a=a)
        calc = GPAW(mode='pw',
                    txt=self.folder / 'bse_al.txt',
                    kpts={'size': (4, 4, 4), 'gamma': True},
                    xc='LDA',
                    nbands=4,
                    convergence={'bands': 'all'})

        atoms.calc = calc
        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def ag_plusU_pw(self):
        xc = 'LDA'
        kpts = 2
        nbands = 6
        pw = 300
        occw = 0.01
        conv = {'bands': nbands,
                'density': 1e-12}
        a = 4.07
        atoms = bulk('Ag', 'fcc', a=a)
        atoms.center()

        atoms.calc = GPAW(
            xc=xc,
            mode=PW(pw),
            kpts={'size': (kpts, kpts, kpts), 'gamma': True},
            setups={'Ag': '11:d,2.0,0'},
            nbands=nbands,
            occupations=FermiDirac(occw),
            convergence=conv,
            parallel={'domain': 1},
            txt=self.folder / 'ag_plusU_pw.txt')

        atoms.get_potential_energy()

        atoms.calc.diagonalize_full_hamiltonian()

        return atoms.calc

    @gpwfile
    def gaas_pw_nosym(self):
        return self._gaas(symmetry='off')

    @gpwfile
    def gaas_pw(self):
        return self._gaas()

    @with_band_cutoff(gpw='gaas_pw',
                      band_cutoff=8)
    def _gaas(self, *, band_cutoff, symmetry=None):
        if symmetry is None:
            symmetry = {}
        nk = 4
        cell = bulk('Ga', 'fcc', a=5.68).cell
        atoms = Atoms('GaAs', cell=cell, pbc=True,
                      scaled_positions=((0, 0, 0), (0.25, 0.25, 0.25)))
        tag = '_nosym' if symmetry == 'off' else ''
        conv = {'bands': band_cutoff + 1,
                'density': 1.e-8}

        calc = GPAW(mode=PW(400),
                    xc='LDA',
                    occupations=FermiDirac(width=0.01),
                    convergence=conv,
                    nbands=band_cutoff + 1,
                    kpts={'size': (nk, nk, nk), 'gamma': True},
                    txt=self.folder / f'gs_GaAs{tag}.txt',
                    symmetry=symmetry)

        atoms.calc = calc
        atoms.get_potential_energy()
        return atoms.calc

    @gpwfile
    def h_pw280_fulldiag(self):
        return self._pw_280_fulldiag(Atoms('H'), hund=True, nbands=4)

    @gpwfile
    def h2_pw280_fulldiag(self):
        return self._pw_280_fulldiag(
            Atoms('H2', [(0, 0, 0), (0, 0, 0.7413)]), nbands=8)

    def _pw_280_fulldiag(self, atoms, **kwargs):
        atoms.set_pbc(True)
        atoms.set_cell((2., 2., 3.))
        atoms.center()
        calc = GPAW(mode=PW(280, force_complex_dtype=True),
                    txt=self.folder / f'{atoms.symbols}_pw_280_fulldiag.txt',
                    xc='LDA',
                    basis='dzp',
                    parallel={'domain': 1},
                    convergence={'density': 1.e-6},
                    **kwargs)
        atoms.calc = calc
        atoms.get_potential_energy()
        calc.diagonalize_full_hamiltonian(nbands=80)
        return calc

    @gpwfile
    def fe_pw_distorted(self):
        xc = 'revTPSS'
        m = [2.9]
        fe = bulk('Fe')
        fe.set_initial_magnetic_moments(m)
        k = 3
        fe.calc = GPAW(mode=PW(800),
                       h=0.15,
                       occupations=FermiDirac(width=0.03),
                       xc=xc,
                       kpts=(k, k, k),
                       convergence={'energy': 1e-8},
                       parallel={'domain': 1, 'augment_grids': True},
                       txt=self.folder / 'fe_pw_distorted.txt')
        fe.set_cell(np.dot(fe.cell,
                           [[1.02, 0, 0.03],
                            [0, 0.99, -0.02],
                            [0.2, -0.01, 1.03]]),
                    scale_atoms=True)
        fe.get_potential_energy()
        return fe.calc

    @gpwfile
    def si8_fd(self):
        a = 5.404
        bulk = Atoms(symbols='Si8',
                     scaled_positions=[(0, 0, 0),
                                       (0, 0.5, 0.5),
                                       (0.5, 0, 0.5),
                                       (0.5, 0.5, 0),
                                       (0.25, 0.25, 0.25),
                                       (0.25, 0.75, 0.75),
                                       (0.75, 0.25, 0.75),
                                       (0.75, 0.75, 0.25)],
                     pbc=True, cell=(a, a, a))
        n = 20
        calc = GPAW(mode='fd',
                    gpts=(n, n, n),
                    nbands=8 * 3,
                    occupations=FermiDirac(width=0.01),
                    kpts=(1, 1, 1))
        bulk.calc = calc
        bulk.get_potential_energy()

        return bulk.calc

    @gpwfile
    def si_pw_distorted(self):
        xc = 'TPSS'
        si = bulk('Si')
        k = 3
        si.calc = GPAW(mode=PW(250),
                       mixer=Mixer(0.7, 5, 50.0),
                       xc=xc,
                       occupations=FermiDirac(0.01),
                       kpts=(k, k, k),
                       convergence={'energy': 1e-8},
                       parallel={'domain': min(2, world.size)},
                       txt=self.folder / 'si_pw_distorted.txt')
        si.set_cell(np.dot(si.cell,
                           [[1.02, 0, 0.03],
                            [0, 0.99, -0.02],
                            [0.2, -0.01, 1.03]]),
                    scale_atoms=True)
        si.get_potential_energy()
        return si.calc

    @gpwfile
    def IBiTe_pw_monolayer(self):
        # janus material. material parameters obtained from c2db.
        from ase.atoms import Atoms
        IBiTe_positions = np.array([[0, 2.552, 7.802],
                                    [0, 0, 9.872],
                                    [2.210, 1.276, 11.575]])
        IBiTe = Atoms('IBiTe', positions=IBiTe_positions)
        IBiTe.pbc = [True, True, False]
        cell = np.array([[4.4219, 0, 0.0, ],
                         [-2.211, 3.829, 0.0],
                         [0.0, 0.0, 19.5]])
        IBiTe.cell = cell
        calc = GPAW(mode=PW(200),
                    xc='LDA',
                    occupations=FermiDirac(0.01),
                    kpts={'size': (6, 6, 1), 'gamma': True},
                    txt=None)
        IBiTe.calc = calc
        IBiTe.get_potential_energy()
        return IBiTe.calc

    def _intraband(self, spinpol: bool):
        atoms = bulk('Na')
        if spinpol:
            atoms.set_initial_magnetic_moments([[0.1]])
        atoms.calc = GPAW(mode=PW(300),
                          kpts={'size': (8, 8, 8), 'gamma': True},
                          parallel={'band': 1},
                          txt=None)
        atoms.get_potential_energy()
        atoms.calc.diagonalize_full_hamiltonian(nbands=20)
        return atoms.calc

    @gpwfile
    def intraband_spinpaired_fulldiag(self):
        return self._intraband(False)

    @gpwfile
    def intraband_spinpolarized_fulldiag(self):
        return self._intraband(True)


# We add Si fixtures with various symmetries to the GPWFiles namespace
for name, method in si_gpwfiles().items():
    setattr(GPWFiles, name, method)



import pytest
import numpy as np

from gpaw import GPAW, restart


def get_restart_test_values(calc):
    atoms = calc.get_atoms()
    # XXX the forces must be evaluated first because a force evaluation
    # somehow affects (changes) both the energy and the wf when restarting
    # from a gpw file. An issue was created (#1051) documenting this unexpected
    # and rather bizarre behavior.
    f = atoms.get_forces()
    e = atoms.get_potential_energy()
    m = atoms.get_magnetic_moments()
    wf = calc.get_pseudo_wave_function(band=1)
    eig0 = calc.get_eigenvalues(spin=0)
    eig1 = calc.get_eigenvalues(spin=1)

    return wf, e, f, m, eig0, eig1


@pytest.fixture(params=['na3_pw_restart', 'na3_fd_restart'])
def gpwfile(request, gpw_files):
    return gpw_files[request.param]


@pytest.mark.later
def test_fileio_restart(in_tmp_dir, gpwfile):
    # gpw restart file is written in fixture
    calc = GPAW(gpwfile)

    wf0, e0, f0, m0, eig00, eig01 = get_restart_test_values(calc=calc)

    # Write the restart file(s):
    calc.write('tmp1.gpw')
    calc.write('tmp.gpw', 'all')

    # Try restarting:
    _, calc = restart('tmp.gpw', txt=None)
    wf1, e1, f1, m1, eig10, eig11 = get_restart_test_values(calc=calc)

    # compare that the values are absolutely equal
    print(e0, e1)
    assert e0 == pytest.approx(e1, abs=1e-10)
    print(f0, f1)
    for ff0, ff1 in zip(f0, f1):
        err = np.linalg.norm(ff0 - ff1)
        assert err <= 1e-10
    print(m0, m1)
    for mm0, mm1 in zip(m0, m1):
        assert mm0 == pytest.approx(mm1, abs=1e-10)
    print("A", eig00, eig10)
    for eig0, eig1 in zip(eig00, eig10):
        assert eig0 == pytest.approx(eig1, abs=1e-10)
    print("B", eig01, eig11)
    for eig0, eig1 in zip(eig01, eig11):
        assert eig0 == pytest.approx(eig1, abs=1e-10)
    assert abs(wf1 - wf0).max() == pytest.approx(0, abs=1e-14)

    # Check that after restart, everything is writable
    calc.write("tmp3.gpw")
    calc.write("tmp4.gpw", "all")



import pytest
import numpy as np
from ase.units import Bohr
from ase.build import bulk
from gpaw import GPAW, FermiDirac
from gpaw.response.df import DielectricFunction, read_response_function
from gpaw.test import findpeak


@pytest.mark.dielectricfunction
@pytest.mark.response
@pytest.mark.libxc
def test_response_diamond_absorption(in_tmp_dir):
    a = 6.75 * Bohr
    atoms = bulk('C', 'diamond', a=a)

    calc = GPAW(mode='pw',
                kpts=(3, 3, 3),
                eigensolver='rmm-diis',
                occupations=FermiDirac(0.001))

    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('C.gpw', 'all')

    eM1_ = 9.727
    eM2_ = 9.548
    w0_ = 10.7782
    I0_ = 5.47
    w_ = 10.7532
    I_ = 5.98

    # Test the old interface to the dielectric constant
    df = DielectricFunction('C.gpw', frequencies=(0.,), eta=0.001, ecut=50,
                            hilbert=False)
    eM1, eM2 = df.get_macroscopic_dielectric_constant()
    assert eM1 == pytest.approx(eM1_, abs=0.01)
    assert eM2 == pytest.approx(eM2_, abs=0.01)

    # ----- RPA dielectric function ----- #
    dfcalc = DielectricFunction(
        'C.gpw', eta=0.25, ecut=50,
        frequencies=np.linspace(0, 24., 241), hilbert=False)
    eps = dfcalc.get_literal_dielectric_function()

    # Test the new interface to the dielectric constant
    eM1, eM2 = eps.dielectric_constant()
    assert eM1 == pytest.approx(eM1_, abs=0.01)
    assert eM2 == pytest.approx(eM2_, abs=0.01)

    # Test the macroscopic dielectric function
    omega_w, eps0M_w, epsM_w = eps.macroscopic_dielectric_function().arrays
    w0, I0 = findpeak(omega_w, eps0M_w.imag)
    assert w0 == pytest.approx(w0_, abs=0.01)
    assert I0 / (4 * np.pi) == pytest.approx(I0_, abs=0.1)
    w, I = findpeak(omega_w, epsM_w.imag)
    assert w == pytest.approx(w_, abs=0.01)
    assert I / (4 * np.pi) == pytest.approx(I_, abs=0.1)

    # Test polarizability
    omega_w, a0rpa_w, arpa_w = eps.polarizability().arrays
    w0, I0 = findpeak(omega_w, a0rpa_w.imag)
    assert w0 == pytest.approx(w0_, abs=0.01)
    assert I0 == pytest.approx(I0_, abs=0.01)
    w, I = findpeak(omega_w, arpa_w.imag)
    assert w == pytest.approx(w_, abs=0.01)
    assert I == pytest.approx(I_, abs=0.01)

    # Test that the macroscopic dielectric function can be calculated also from
    # the inverse dielectric function and the bare dielectric function
    epsinv = dfcalc.get_inverse_dielectric_function()
    _, _, epsM_frominv_w = epsinv.macroscopic_dielectric_function().arrays
    assert epsM_frominv_w == pytest.approx(epsM_w, rel=1e-6)
    epsbare = dfcalc.get_bare_dielectric_function()
    _, _, epsM_frombare_w = epsbare.macroscopic_dielectric_function().arrays
    assert epsM_frombare_w == pytest.approx(epsM_w, rel=1e-6)

    # ----- TDDFT absorption spectra ----- #

    # Absorption spectrum calculation ALDA
    w_ = 10.7562
    I_ = 5.8803

    epsinv = dfcalc.get_inverse_dielectric_function(xc='ALDA', rshelmax=0)
    # Here we base the check on a written results file
    epsinv.polarizability().write(filename='ALDA_pol.csv')
    dfcalc.context.comm.barrier()
    omega_w, a0alda_w, aalda_w = read_response_function('ALDA_pol.csv')

    assert a0alda_w == pytest.approx(a0rpa_w, rel=1e-4)
    w, I = findpeak(omega_w, aalda_w.imag)
    assert w == pytest.approx(w_, abs=0.01)
    assert I == pytest.approx(I_, abs=0.1)

    # Absorption spectrum calculation long-range kernel
    w_ = 10.2906
    I_ = 5.6955

    epsinv = dfcalc.get_inverse_dielectric_function(xc='LR0.25')
    omega_w, a0lr_w, alr_w = epsinv.polarizability().arrays

    assert a0lr_w == pytest.approx(a0rpa_w, rel=1e-4)
    w, I = findpeak(omega_w, alr_w.imag)
    assert w == pytest.approx(w_, abs=0.01)
    assert I == pytest.approx(I_, abs=0.1)

    # Absorption spectrum calculation Bootstrap
    w_ = 10.4600
    I_ = 6.0263

    epsinv = dfcalc.get_inverse_dielectric_function(xc='Bootstrap')
    omega_w, a0btsr_w, abtsr_w = epsinv.polarizability().arrays

    assert a0btsr_w == pytest.approx(a0rpa_w, rel=1e-4)
    w, I = findpeak(omega_w, abtsr_w.imag)
    assert w == pytest.approx(w_, abs=0.02)
    assert I == pytest.approx(I_, abs=0.2)

    # import matplotlib.pyplot as plt
    # plt.plot(omega_w, a0rpa_w.imag, label='IP')
    # plt.plot(omega_w, arpa_w.imag, label='RPA')
    # plt.plot(omega_w, aalda_w.imag, label='ALDA')
    # plt.plot(omega_w, alr_w.imag, label='LR0.25')
    # plt.plot(omega_w, abtsr_w.imag, label='Bootstrap')
    # plt.legend()
    # plt.show()



import numpy as np
import pytest
from gpaw import GPAW
from gpaw.response import ResponseContext, ResponseGroundStateAdapter
from gpaw.response.frequencies import FrequencyDescriptor
from gpaw.response.modelinteraction import initialize_w_model
from gpaw.response.chi0 import Chi0Calculator
from gpaw.wannier90 import Wannier90
import os
from gpaw.mpi import world, serial_comm
from subprocess import PIPE, run


def out():
    result = run('wannier90.x --version',
                 stdout=PIPE,
                 stderr=PIPE,
                 universal_newlines=True,
                 shell=True)
    return result.stdout


@pytest.mark.later
@pytest.mark.parametrize('symm', [True, False])
@pytest.mark.response
@pytest.mark.skipif(': 3.' not in out(),
                    reason="requires at least Wannier90 version 3.0")
def test_w(in_tmp_dir, gpw_files, symm):

    if not symm and world.size < 2:
        pytest.skip('Skip nosymm test in serial')

    if symm:
        gpwfile = gpw_files['gaas_pw']
    else:
        gpwfile = gpw_files['gaas_pw_nosym']

    calc = GPAW(gpwfile, communicator=serial_comm)
    seed = 'GaAs'

    # Wannier90 only works in serial
    if world.rank == 0:
        w90 = Wannier90(calc, orbitals_ai=[[], [0, 1, 2, 3]],
                        bands=range(4),
                        seed=seed)
        w90.write_input(num_iter=1000,
                        plot=True,
                        write_u_matrices=True)
        w90.write_wavefunctions()
        os.system('wannier90.x -pp ' + seed)
        w90.write_projections()
        w90.write_eigenvalues()
        w90.write_overlaps()
        os.system('wannier90.x ' + seed)

    world.barrier()

    omega = np.array([0])
    kwargs = dict(hilbert=False, ecut=30, intraband=False)
    gs = ResponseGroundStateAdapter.from_gpw_file(gpwfile)
    context = ResponseContext('test.log')
    wd = FrequencyDescriptor.from_array_or_dict(omega)
    chi0calc = Chi0Calculator(gs, context, wd=wd, **kwargs)
    Wm = initialize_w_model(chi0calc)
    w, Wwann = Wm.calc_in_Wannier(chi0calc, Uwan_mnk=seed, bandrange=[0, 4])
    check_W(Wwann)
    assert np.allclose(w, omega)

    # test block parallelization
    if world.size % 2 == 0 and symm:
        omega = np.array([0, 1])
        wd = FrequencyDescriptor.from_array_or_dict(omega)
        chi0calc = Chi0Calculator(gs, context, wd=wd, nblocks=2, **kwargs)
        Wm = initialize_w_model(chi0calc)
        w, Wwann = Wm.calc_in_Wannier(chi0calc, Uwan_mnk=seed,
                                      bandrange=[0, 4])
        check_W(Wwann)
        print(omega, w)
        assert np.allclose(w, omega)


def check_W(Wwann):
    assert Wwann[0, 0, 0, 0, 0] == pytest.approx(2.478, abs=0.003)
    assert Wwann[0, 1, 1, 1, 1] == pytest.approx(1.681, abs=0.003)
    assert Wwann[0, 2, 2, 2, 2] == pytest.approx(1.681, abs=0.003)
    assert Wwann[0, 3, 3, 3, 3] == pytest.approx(1.681, abs=0.003)
    assert Wwann[0, 3, 3, 0, 0] == pytest.approx(0.861, abs=0.003)
    assert Wwann[0, 3, 0, 3, 0].real == pytest.approx(1.757, abs=0.003)
    assert np.abs(Wwann[0, 3, 0, 3, 0].imag) < 0.005



import numpy as np

from ase.units import Bohr
from ase.neighborlist import natural_cutoffs, build_neighbor_list

from gpaw.sphere.integrate import (integrate_lebedev,
                                   radial_truncation_function,
                                   spherical_truncation_function_collection,
                                   default_spherical_drcut,
                                   find_volume_conserving_lambd)
from gpaw.response import ResponseGroundStateAdapter
from gpaw.response.localft import (add_spin_polarization,
                                   add_LSDA_zeeman_energy)


class AtomicSites:
    """Object defining a set of spherical atomic sites."""

    def __init__(self, indices, radii):
        """Construct the AtomicSites.

        Parameters
        ----------
        indices : 1D array-like
            Atomic index A for each site index a.
        radii : 2D array-like
            Atomic radius rc for each site index a and partitioning p.
        """
        self.A_a = np.asarray(indices)
        assert self.A_a.ndim == 1
        assert len(np.unique(self.A_a)) == len(self.A_a)

        # Parse the input atomic radii
        rc_ap = np.asarray(radii)
        assert rc_ap.ndim == 2
        assert rc_ap.shape[0] == len(self.A_a)
        # Convert radii to internal units (Å to Bohr)
        self.rc_ap = rc_ap / Bohr

        self.npartitions = self.rc_ap.shape[1]
        self.shape = rc_ap.shape

    def __len__(self):
        return len(self.A_a)


class AtomicSiteData:
    r"""Data object for a set of spherical atomic sites."""

    def __init__(self, gs: ResponseGroundStateAdapter, sites: AtomicSites):
        """Extract atomic site data from a ground state adapter."""
        assert self._in_valid_site_radii_range(gs, sites), \
            'Please provide site radii in the valid range, see '\
            'AtomicSiteData.valid_site_radii_range()'
        self.sites = sites

        # Extract the scaled positions and micro_setups for each atomic site
        self.spos_ac = gs.spos_ac[sites.A_a]
        self.micro_setup_a = [gs.micro_setups[A] for A in sites.A_a]

        # Extract pseudo density on the fine real-space grid
        self.finegd = gs.finegd
        self.nt_sr = gs.nt_sr

        # Set up the atomic truncation functions which define the sites based
        # on the coarse real-space grid
        self.gd = gs.gd
        self.drcut = default_spherical_drcut(self.gd)
        self.lambd_ap = np.array(
            [[find_volume_conserving_lambd(rcut, self.drcut)
              for rcut in rc_p] for rc_p in sites.rc_ap])
        self.stfc = spherical_truncation_function_collection(
            self.finegd, self.spos_ac, sites.rc_ap, self.drcut, self.lambd_ap)

    @staticmethod
    def _valid_site_radii_range(gs):
        """For each atom in gs, determine the valid site radii range in Bohr.

        The lower bound is determined by the spherical truncation width, when
        truncating integrals on the real-space grid.
        The upper bound is determined by the distance to the nearest
        augmentation sphere.
        """
        atoms = gs.atoms
        drcut = default_spherical_drcut(gs.gd)
        rmin_A = np.array([drcut / 2] * len(atoms))

        # Find neighbours based on covalent radii
        cutoffs = natural_cutoffs(atoms, mult=2)
        neighbourlist = build_neighbor_list(atoms, cutoffs,
                                            self_interaction=False)
        # Determine rmax for each atom
        augr_A = gs.get_aug_radii()
        rmax_A = []
        for A in range(len(atoms)):
            pos = atoms.positions[A]
            # Calculate the distance to the augmentation sphere of each
            # neighbour
            aug_distances = []
            for An, offset in zip(*neighbourlist.get_neighbors(A)):
                posn = atoms.positions[An] + offset @ atoms.get_cell()
                dist = np.linalg.norm(posn - pos) / Bohr  # Å -> Bohr
                aug_dist = dist - augr_A[An]
                assert aug_dist > 0.
                aug_distances.append(aug_dist)
            # In order for PAW corrections to be valid, we need a sphere of
            # radius rcut not to overlap with any neighbouring augmentation
            # spheres
            rmax_A.append(min(aug_distances))
        rmax_A = np.array(rmax_A)

        return rmin_A, rmax_A

    @staticmethod
    def valid_site_radii_range(gs):
        """Get the valid site radii for all atoms in a given ground state."""
        rmin_A, rmax_A = AtomicSiteData._valid_site_radii_range(gs)
        # Convert to external units (Bohr to Å)
        return rmin_A * Bohr, rmax_A * Bohr

    @staticmethod
    def _in_valid_site_radii_range(gs, sites):
        rmin_A, rmax_A = AtomicSiteData._valid_site_radii_range(gs)
        for a, A in enumerate(sites.A_a):
            if not np.all(
                    np.logical_and(
                        sites.rc_ap[a] > rmin_A[A] - 1e-8,
                        sites.rc_ap[a] < rmax_A[A] + 1e-8)):
                return False
        return True

    def calculate_magnetic_moments(self):
        """Calculate the magnetic moments at each atomic site."""
        magmom_ap = self.integrate_local_function(add_spin_polarization)
        return magmom_ap

    def calculate_zeeman_energies(self):
        r"""Calculate the local Zeeman energy E_Z for each atomic site."""
        EZ_ap = self.integrate_local_function(add_LSDA_zeeman_energy)
        return EZ_ap

    def integrate_local_function(self, add_f):
        r"""Integrate a local function f[n](r) = f(n(r)) over the atomic sites.

        For every site index a and partitioning p, the integral is defined via
        a smooth truncation function θ(|r-r_a|<rc_ap):

               /
        f_ap = | dr θ(|r-r_a|<rc_ap) f(n(r))
               /
        """
        out_ap = np.zeros(self.sites.shape, dtype=float)
        self._integrate_pseudo_contribution(add_f, out_ap)
        self._integrate_paw_correction(add_f, out_ap)
        return out_ap

    def _integrate_pseudo_contribution(self, add_f, out_ap):
        """Calculate the pseudo contribution to the atomic site integrals.

        For local functions of the density, the pseudo contribution is
        evaluated by a numerical integration on the real-space grid:

        ̰       /
        f_ap = | dr θ(|r-r_a|<rc_ap) f(ñ(r))
               /
        """
        # Evaluate the local function on the real-space grid
        ft_r = self.finegd.zeros()
        add_f(self.finegd, self.nt_sr, ft_r)

        # Integrate θ(|r-r_a|<rc_ap) f(ñ(r))
        ftdict_ap = {a: np.empty(self.sites.npartitions)
                     for a in range(len(self.sites))}
        self.stfc.integrate(ft_r, ftdict_ap)

        # Add pseudo contribution to the output array
        for a in range(len(self.sites)):
            out_ap[a] += ftdict_ap[a]

    def _integrate_paw_correction(self, add_f, out_ap):
        """Calculate the PAW correction to an atomic site integral.

        The PAW correction is evaluated on the atom centered radial grid, using
        the all-electron and pseudo densities generated from the partial waves:

                /
        Δf_ap = | r^2 dr θ(r<rc_ap) [f(n_a(r)) - f(ñ_a(r))]
                /
        """
        for a, (micro_setup, rc_p, lambd_p) in enumerate(zip(
                self.micro_setup_a, self.sites.rc_ap, self.lambd_ap)):
            # Evaluate the PAW correction and integrate angular components
            df_ng = micro_setup.evaluate_paw_correction(add_f)
            df_g = integrate_lebedev(df_ng)
            for p, (rcut, lambd) in enumerate(zip(rc_p, lambd_p)):
                # Evaluate the smooth truncation function
                theta_g = radial_truncation_function(
                    micro_setup.rgd.r_g, rcut, self.drcut, lambd)
                # Integrate θ(r) Δf(r) on the radial grid
                out_ap[a, p] += micro_setup.rgd.integrate_trapz(df_g * theta_g)
